/*	$LAAS: correlation.tmpl.c 2010/12/22 17:32:19 mallet $ */

/*
 * Copyright (c) 2000,2008,2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                                        Fri Aug  4 2006
 */

#include "stereopixel-config.h"

/* check that the template variables are correctly defined */
#ifndef SPIX_CORREL_NAME
# error	"Correlation algorithm name is not defined"
#endif
#ifndef SPIX_CORREL_COMBINE_TYPE
# error	"Pixel combination type is not defined"
#endif
#ifndef SPIX_CORREL_SUM_TYPE
# error	"Combined pixel sum type is not defined"
#endif
#ifndef SPIX_CORREL_SCORE_TYPE
# error	"Correlation score type is not defined"
#endif

#include <stdio.h>
#include <err.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "libstereopixel.h"
#include "vectorization.h"
#if SPIX_THREADS
# include "workers.h"
#endif


/* --- local data ---------------------------------------------------- */

#define spix_glue(x, y)		x ## y
#define spix_func_name(x, y)	spix_glue(x,y)

#define MIN(a,b)		((a)<(b)?(a):(b))

static inline int
		spix_correl_chkparam(unsigned char *source_gauche,
			unsigned char *source_droite,
			SP_IMAGES_TMP *images_gauches,
			SP_IMAGES_TMP *images_droites,
			SP_CORRELATION_PARAMS *correlation_params,
			SP_CORRELATION_RESULTS *correlation_results);
static inline int
		spix_correl_preproc(unsigned char **source_gauche,
			unsigned char **source_droite,
			SP_IMAGES_TMP *images_gauches, SP_IMAGES_TMP *images_droites);
static inline SPIX_CORREL_SUM_TYPE
		sum_products_window(SPIX_CORREL_COMBINE_TYPE **plines, int j,
			int ni, int nj, int d, int disp_min_general,
			int delta_max_disp)
		__attribute__ ((pure));

static int	spix_func_name(spix_docorrel_,SPIX_CORREL_NAME)(
			unsigned char *source_gauche,
			unsigned char *source_droite,
			SP_IMAGES_TMP *images_gauches,
			SP_IMAGES_TMP *images_droites,
			void **lines_products, void *sum_products_top,
 			void *sum_products_bot, void *mem_scores,
			void *mem_scores_2, int *mem_max,
			SP_CORRELATION_PARAMS *correlation_params,
			SP_CORRELATION_RESULTS *correlation_results, int i_min,
			int i_max);
static void	spix_reversecheck(float *pix_disp, float *pix_sigma,
			float *pix_sigma_g, int *tab_max,
			SPIX_CORREL_SCORE_TYPE *tab_scores,
			SP_CORRELATION_PARAMS *correlation_params,
			SP_CORRELATION_RESULTS *correlation_results, int nblig,
			int j_min, int j_max);

static inline SPIX_CORREL_COMBINE_TYPE
		spix_correl_pixel_combine(unsigned char a, unsigned char b)
		__attribute__ ((pure));
static inline void
		spix_correl_pixel_combine_vec(SPIX_CORREL_COMBINE_TYPE *c,
			spix_vector(unsigned char) a,
			spix_vector(unsigned char) b);
static inline SPIX_CORREL_SCORE_TYPE
		spix_correl_score(SPIX_CORREL_SUM_TYPE sum, double norm,
			float lavg, float ravg, float lrstddev,
			float rrstddev)
		__attribute__ ((pure));
static inline spix_vector(SPIX_CORREL_SCORE_TYPE)
		spix_correl_score_vec(spix_vector(SPIX_CORREL_SUM_TYPE) sum,
			double norm, spix_vector(float) lavg,
			spix_vector(float) ravg, spix_vector(float) lrstddev,
			spix_vector(float) rrstddev)
		__attribute__ ((pure));



/* --- spix_correl_{zncc,census} ------------------------------------- */

/** Perform stereo correlation
 */

int
spix_func_name(spix_correl_,SPIX_CORREL_NAME)(
  unsigned char *source_gauche, unsigned char *source_droite,
  SP_IMAGES_TMP *images_gauches, SP_IMAGES_TMP *images_droites,
  SP_CORRELATION_PARAMS *correlation_params,
  SP_CORRELATION_RESULTS *correlation_results)
{
  struct rusage start, end;

  void **tab_tab_produits = correlation_results->tab_tab_produits;
  void *sum_products_top = correlation_results->somme_pdt_haut;
  void *sum_products_bot = correlation_results->somme_pdt_bas;
  void *tab_scores_2 = correlation_results->tab_scores_2;
  void *tab_scores = correlation_results->tab_scores;
  int *tab_max = correlation_results->tab_max;

  int nblig = images_droites->images_sizes.nb_lig_final;
  int nbcol = images_droites->images_sizes.nb_col_final;
  int ni = correlation_params->win_nb_lig;
  int i_min = ni/2;
  int i_max = nblig - ni/2;
  int t, s;

  int nbloop;

  i_max--;

  /* check input parameters */
  s = spix_correl_chkparam(source_gauche, source_droite,
			   images_gauches, images_droites,
			   correlation_params, correlation_results);
  if (s) return s;

  if (correlation_params->state != SP_CORRELATION_PARAMS_INITIALIZED)
    return stereoError(ERR_SP_CORRELATION_PARAMS_NOT_INITIALIZED);

  if (correlation_results->state != SP_CORRELATION_RESULTS_ALLOCATED)
    return stereoError(ERR_SP_CORRELATION_RESULTS_NOT_ALLOCATED);

  getrusage(RUSAGE_SELF, &start);

  /* apply preprocessings */
  s = spix_correl_preproc(&source_gauche, &source_droite,
			  images_gauches, images_droites);
  if (s) return s;

  /* init disparity and stddev images */
  memset(correlation_results->im_disparites, 0, nblig * nbcol * sizeof(float));

  if (correlation_params->type_correlation == SP_CORRELATION_ZNCC_SIGMA)
    memset(correlation_results->im_sigma_disparites, 0,
	   nblig * nbcol * sizeof(float));

#if SPIX_THREADS
  if (correlation_params->threads > 1) {
    spix_worker *w;

    for(t = 0; t < correlation_params->threads-1; t++) {
      w = correlation_params->thread_array[t];

      w->source_gauche = source_gauche;
      w->source_droite = source_droite;
      w->images_gauches = images_gauches;
      w->images_droites = images_droites;
      w->correlation_results = correlation_results;
      w->job = spix_func_name(spix_docorrel_,SPIX_CORREL_NAME);
    }

    w = correlation_params->thread_array[0];
    i_max = w->i_min;
    pthread_barrier_destroy(w->done);
    pthread_barrier_init(w->done, NULL, correlation_params->threads);
    pthread_barrier_wait(w->start);
  }

  s = spix_func_name(spix_docorrel_,SPIX_CORREL_NAME)(
    source_gauche, source_droite, images_gauches, images_droites,
    tab_tab_produits, sum_products_top, sum_products_bot, tab_scores,
    tab_scores_2, tab_max, correlation_params, correlation_results, i_min,
    i_max);

  if (correlation_params->threads > 1) {
    spix_worker *w = correlation_params->thread_array[0];

    pthread_barrier_destroy(w->reset);
    pthread_barrier_init(w->reset, NULL, correlation_params->threads);
    pthread_barrier_wait(w->done);
    pthread_barrier_destroy(w->start);
    pthread_barrier_init(w->start, NULL, correlation_params->threads);
    pthread_barrier_wait(w->reset);
  }
#else
  s = spix_func_name(spix_docorrel_,SPIX_CORREL_NAME)(
    source_gauche, source_droite, images_gauches, images_droites,
    tab_tab_produits, sum_products_top, sum_products_bot, tab_scores,
    tab_scores_2, tab_max, correlation_params,
    correlation_results, i_min, i_max);
#endif /* SPIX_THREADS */

  correlation_results->state_disparites = SP_IMAGE_COMPUTED;
  if (correlation_params->type_correlation == SP_CORRELATION_ZNCC_SIGMA)
    correlation_results->state_sigma_disparites = SP_IMAGE_COMPUTED;

  getrusage(RUSAGE_SELF, &end);
  warnx("%gms user, %gms sys",
	 end.ru_utime.tv_sec*1000. + end.ru_utime.tv_usec/1000. -
	 start.ru_utime.tv_sec*1000. - start.ru_utime.tv_usec/1000.,
	 end.ru_stime.tv_sec*1000. + end.ru_stime.tv_usec/1000. -
	 start.ru_stime.tv_sec*1000. - start.ru_stime.tv_usec/1000.);
  warnx("removed %d pixels", correlation_results->nbRemovedByUnicity);
  return s;
}


/* --- spix_docorrel_{zncc,census} ----------------------------------- */

static int
spix_func_name(spix_docorrel_,SPIX_CORREL_NAME)(
  unsigned char *source_gauche, unsigned char *source_droite,
  SP_IMAGES_TMP *images_gauches, SP_IMAGES_TMP *images_droites,
  void **lines_products, void *sum_products_top, void *sum_products_bot,
  void *mem_scores, void *mem_scores_2, int *mem_max,
  SP_CORRELATION_PARAMS *correlation_params,
  SP_CORRELATION_RESULTS *correlation_results, int i_min, int i_max)
{
  /* Les paramtres inclus dans les structures */
  unsigned char* q_im_g;
  unsigned char* q_im_d;
  float*         q_im_sigma_g;
  float*         q_im_sigma_d;
  float*         q_im_somme_g;
  float*         q_im_somme_d;

  float*	q_im_disp;
  float*	q_im_sigma;

  int            nblig;
  int            nbcol;

  int            ni;
  int            nj;
  int            disparite_min;
  int            disparite_max;
  SPIX_CORREL_SCORE_TYPE max_score;

  double         seuil_sigma;
  double         seuil_score;
  double         seuil_voisins;
  int            epsilon_inverse;

  /* Pixels des diffrentes images */
  float*         pix_sigma_g;
  float*         pix_sigma_g_bis;
  float*         pix_sigma_d;
  float*         pix_somme_g;
  float*         pix_somme_d;
  float*         pix_disp;
  float*         pix_sigma;

  /* COMMENTER MIEUX L'ENSEMBLE DES VARIABLES QUI SUIVENT !!! */

  /* Les bornes sur lesquelles il faut corrler */
  int             j_min, j_max;

  int             ni2, nj2;	/* ni / 2 et nj / 2 */

  unsigned char*  pix_haut_d;
  unsigned char*  pix_haut_g;
  unsigned char*  pix_bas_d;
  unsigned char*  pix_bas_g;

  spix_vector(unsigned char) pixel;

  /* Les diffrents tableaux intermdiaires */

  /* Les tableaux de produits : de taille (nbcol * delta_max_disp), ils
     contiennent les produits des pixels (j) de l'image de gauche avec les
     pixels (j + d) de l'images de droite, pour tout j et tout d. On en stocke
     "ni", et on utilise les premier et le dernier pour dduire les scores de
     corrlation d'un ligne I de ceux calculs en une ligne I - 1. */

  SPIX_CORREL_COMBINE_TYPE *tab_produits_haut;
  SPIX_CORREL_COMBINE_TYPE *tab_produits_haut_ptr;
  SPIX_CORREL_COMBINE_TYPE *pix_pdt_haut_gauche;
  SPIX_CORREL_COMBINE_TYPE *pix_pdt_haut_droite;

  SPIX_CORREL_COMBINE_TYPE *tab_produits_bas;
  SPIX_CORREL_COMBINE_TYPE *tab_produits_bas_ptr;
  SPIX_CORREL_COMBINE_TYPE *pix_pdt_bas_gauche;
  SPIX_CORREL_COMBINE_TYPE *pix_pdt_bas_droite;

  SPIX_CORREL_COMBINE_TYPE *tab_produits;
  SPIX_CORREL_COMBINE_TYPE *tab_produits_ptr;

  SPIX_CORREL_COMBINE_TYPE **tab_tab_produits;


  /* Les tableaux "somme_pdt" : de taille (delta_max_disp), ils servent 
     stocker les sommes de "nj" produits des "tab_produits" */
  SPIX_CORREL_SUM_TYPE *somme_pdt_bas;
  SPIX_CORREL_SUM_TYPE *somme_pdt_bas_ptr;

  SPIX_CORREL_SUM_TYPE *somme_pdt_haut;
  SPIX_CORREL_SUM_TYPE *somme_pdt_haut_ptr;

  /* Quelques petites variables pour initialiser les tableaux prcedents */
  SPIX_CORREL_SUM_TYPE somme_col_droite;
  SPIX_CORREL_SUM_TYPE somme_col_gauche;
  SPIX_CORREL_SUM_TYPE psum;

  /* les scores : dans un tableau de taille (nbcol * delta_max_disp) */
  int *tab_max;
  SPIX_CORREL_SUM_TYPE *tab_scores;
  SPIX_CORREL_SUM_TYPE *tab_scores_ptr;

  SPIX_CORREL_SCORE_TYPE *tab_scores_2;
  SPIX_CORREL_SCORE_TYPE *tab_scores_2_ptr;

  /* Les intervalles de disparit */
  int*      tab_disp_min;
  int       disp_min;
  int*      tab_disp_max;
  int       disp_max;
  int*      tab_disp_nj2_min;
  int*      tab_disp_nj2_max;
				/* tableaux de taille "nb_colonnes" o
				   on met les disparits min et max
				   calcules pour chaque colonne */

  int       delta_max_disp;
  int       disp_min_general;

  /* Variables pour boucler */
  int       i, j, k, l, d;

 /* Variables d'optimisation : _P_ dsigne "*" (Produit), _S_ dsigne "+"
     (Somme) et _D_ dsigne "-" (Diffrence) */
  int       i_S_ni2_P_nbcol_g;  /* ( (i + ni2) * nbcol_g ) */
  double    prod_taille_inverse ; /* 1 / (ni * nj) : pour remplacer une
				    division par une multiplication dans les
				    calcul des scores */

  /* Encore un paquet de variables d'optimisation ajoutes plus tard */
  unsigned int index;		/* Utilise comme indice commun 
				   tab_scores_ptr et tab_scores_2_ptr */
  spix_vector(float) vec_suml, vec_sigl;
  spix_vector(SPIX_CORREL_SCORE_TYPE) vec_max_score;
  spix_vector(SPIX_CORREL_SCORE_TYPE) vec_max_ind;
  spix_vector(SPIX_CORREL_SCORE_TYPE) vec_ind;

  /* get and initialize input parameters and local variables */
  q_im_g = source_gauche;
  q_im_sigma_g = images_gauches->im_sigmas;
  q_im_somme_g = images_gauches->im_moyennes;

  q_im_d = source_droite;
  q_im_sigma_d = images_droites->im_sigmas;
  q_im_somme_d = images_droites->im_moyennes;

  ni = correlation_params->win_nb_lig;
  nj = correlation_params->win_nb_col;
  ni2 = ni / 2;
  nj2 = nj / 2;
  prod_taille_inverse = 1.0 / (ni * nj);

  nblig = images_droites->images_sizes.nb_lig_final;
  nbcol = images_droites->images_sizes.nb_col_final;
  j_max = nbcol - nj2 ;
  j_min = nj2;

  q_im_disp = correlation_results->im_disparites;

  q_im_sigma = correlation_results->im_sigma_disparites;

  disparite_min = correlation_params->disp_min;
  disparite_max = correlation_params->disp_max;

  seuil_sigma = correlation_params->var_th;
  seuil_score = correlation_params->score_th;
  seuil_voisins = correlation_params->diff_neighbor_max_th;
  epsilon_inverse = correlation_params->inv_th;

  delta_max_disp = correlation_results->delta_max_disp;
  disp_min_general = correlation_results->disp_min_general;

  tab_disp_min = correlation_results->tab_disp_min;
  tab_disp_max = correlation_results->tab_disp_max;
  tab_disp_nj2_min = correlation_results->tab_disp_nj2_min;
  tab_disp_nj2_max = correlation_results->tab_disp_nj2_max;

  tab_max = mem_max;
  tab_scores_2 = mem_scores_2;
  tab_scores = mem_scores;
  somme_pdt_bas = sum_products_bot;
  somme_pdt_haut = sum_products_top;
  tab_tab_produits = (SPIX_CORREL_COMBINE_TYPE **)lines_products;
  tab_produits_bas = tab_tab_produits[ni];

  memset(tab_produits_bas, 0, delta_max_disp * nbcol * sizeof(SPIX_CORREL_COMBINE_TYPE));


  /* -----------------------------------------------------------------------
   *
   * Initialize the arrays of pixel-by-pixel combinations over the first ni
   * lines and the full width of the images.
   */

  for(i = 0; i < ni; i++) {

    tab_produits = *(tab_tab_produits + i);	/* ith line */
    pix_haut_g = q_im_g + (i+i_min-ni2) * nbcol;/* (0, i) in the left image */

    for(j = 0; j < nbcol; j++) {

      disp_min = tab_disp_min[j];
      disp_max = tab_disp_max[j];

      tab_produits_ptr =		/* (j, disp_min - disp_min_general) */
	tab_produits + (j * delta_max_disp) + (disp_min - disp_min_general);

					/* (j + disp_min, i) in right image */
      pix_haut_d = q_im_d + (i+i_min-ni2) * nbcol + j + disp_min;

      pixel = spix_1char(*pix_haut_g);
      for(d = disp_max - disp_min; d >= 0; d-=spix_vector_len(char)) {
	spix_correl_pixel_combine_vec(
	  tab_produits_ptr, spix_ldu(char, pix_haut_d), pixel);
	tab_produits_ptr += spix_vector_len(char);
	pix_haut_d += spix_vector_len(char);
      }

      pix_haut_g++;		/* shift one pixel right in left image */
    }
  }


  /* -----------------------------------------------------------------------
   *
   * Compute scores for the first pixel (j_min, i_min) over all the disparities
   */

  memset(tab_scores, 0, delta_max_disp * nbcol * sizeof(SPIX_CORREL_SUM_TYPE));
  memset(tab_scores_2, 0, delta_max_disp * nbcol * sizeof(SPIX_CORREL_SCORE_TYPE));
  memset(tab_max, 0, nbcol * sizeof(int));

  disp_min = tab_disp_nj2_min[j_min];
  disp_max = tab_disp_nj2_max[j_min];

  pix_somme_g = q_im_somme_g + i_min * nbcol + j_min;
  pix_somme_d = q_im_somme_d + i_min * nbcol + j_min + disp_min;

  pix_sigma_g = q_im_sigma_g + i_min * nbcol + j_min;
  pix_sigma_d = q_im_sigma_d + i_min * nbcol + j_min + disp_min;

  max_score = -1;
  for(d = disp_min; d <= disp_max; d++) {

    psum = sum_products_window(
      tab_tab_produits, j_min, ni, nj, d, disp_min_general, delta_max_disp);

    index = (j_min * delta_max_disp) + (d - disp_min_general);
    tab_scores[index] = psum;
    tab_scores_2[index] =
      spix_correl_score(psum, prod_taille_inverse, *pix_somme_g, *pix_somme_d,
			*pix_sigma_d, *pix_sigma_g);

    if (tab_scores_2[index] > max_score) {
      max_score = tab_scores_2[index];
      tab_max[j_min] = d;
    }

    pix_somme_d++;
    pix_sigma_d++;
  }


  /* -----------------------------------------------------------------------
   *
   * Compute scores for the remaining pixels of the first line, over all the
   * disparities.
   */

  for(j = j_min + 1; j < j_max; j++) {

    disp_min = tab_disp_nj2_min[j];
    disp_max = tab_disp_nj2_max[j];

    pix_somme_g = q_im_somme_g + j + i_min * nbcol;
    pix_somme_d = q_im_somme_d + (j + disp_min) + i_min * nbcol;

    pix_sigma_g = q_im_sigma_g + j + i * nbcol;
    pix_sigma_d = q_im_sigma_d + (j + disp_min) + i_min * nbcol;

    /* the score for the minimum disparity first */
    psum = sum_products_window(
      tab_tab_produits, j, ni, nj, disp_min, disp_min_general, delta_max_disp);

    index = (j * delta_max_disp) + (disp_min - disp_min_general);
    tab_scores[index] =  psum;
    tab_scores_2[index] =
      spix_correl_score(psum, prod_taille_inverse, *pix_somme_g, *pix_somme_d,
			*pix_sigma_d, *pix_sigma_g);

    max_score = tab_scores_2[index];
    tab_max[j] = disp_min;

    pix_somme_d++;
    pix_sigma_d++;

    /* iterate over the remaining disparities: the score of pixel (j,i) is
     * computed from the score of pixel (j-1,i) by summing the left
     * (j -nj/2-1) and right (j + nj2) columns of the pixel-by-pixel
     * combinations. */

    for(d = disp_min + 1; d <= disp_max; d++) {

      somme_col_droite = 0;
      somme_col_gauche = 0;
      for(l = 0; l < ni; l++) {
	tab_produits = tab_tab_produits[l];
	somme_col_gauche
	  += *(tab_produits + ( (j - nj2 - 1) * delta_max_disp)
	       + (d - disp_min_general));
	somme_col_droite
	  += *(tab_produits + ( (j + nj2) * delta_max_disp)
	       + (d - disp_min_general));
      }

      index = (j * delta_max_disp) + (d - disp_min_general);
      tab_scores[index] =
	tab_scores[index-delta_max_disp] + somme_col_droite-somme_col_gauche;

      tab_scores_2[index] =
	spix_correl_score(tab_scores[index], prod_taille_inverse, *pix_somme_g,
			  *pix_somme_d, *pix_sigma_d, *pix_sigma_g);

      if (tab_scores_2[index] > max_score) {
	max_score = tab_scores_2[index];
	tab_max[j] = d;
      }

      pix_somme_d++;
      pix_sigma_d++;
    }
  }

  /* On se positionne  la colonne en (i_min, j_min) dans l'image de disparit,
     l'image des sigmas gauche et l'image des carts types sur les
     disparits */
  pix_disp = q_im_disp + i_min + (j_min * nblig);
  if (correlation_params->type_correlation == SP_CORRELATION_ZNCC_SIGMA ||
      correlation_params->type_correlation == SP_CORRELATION_CENSUS_SIGMA)
    pix_sigma = q_im_sigma + i_min + (j_min * nblig);
  else
    pix_sigma = NULL;
  if (images_gauches->state_sigmas == SP_IMAGE_COMPUTED)
    pix_sigma_g_bis = q_im_sigma_g + i_min * nbcol + j_min;
  else
    pix_sigma_g_bis = NULL;

  correlation_results->nbRemovedByUnicity = 0;
  spix_reversecheck(pix_disp, pix_sigma, pix_sigma_g_bis,
		    tab_max, tab_scores_2,
		    correlation_params, correlation_results,
		    nblig, j_min, j_max);

  pix_disp++;
  if (pix_sigma) pix_sigma++;
  if (pix_sigma_g_bis) pix_sigma_g_bis+=	nbcol;


  /* -----------------------------------------------------------------------
   *
   * Compute the remaining lines.
   */

  i_S_ni2_P_nbcol_g = (i_min + ni2) * nbcol;
  for(i = i_min + 1; i < i_max; i++) {
    i_S_ni2_P_nbcol_g += nbcol;

    /* rotate pixel-by-pixel combinations arrays: tab_produit_haut is line
     * (i-ni2 -1) and tab_produit_bas is line (i+ni2) */
    tab_produits_haut = tab_tab_produits[0];
    tab_produits_bas = tab_tab_produits[ni];
    for (k = 0; k < ni; k++) {
      tab_tab_produits[k] = tab_tab_produits[k + 1];
    }
    tab_tab_produits[ni] = tab_produits_haut;

    /* compute the new line in tab_produits_bas, starting from line i+ni/2 in
     * the left image */
    pix_bas_g = q_im_g + i_S_ni2_P_nbcol_g;

    for(j = 0; j < nbcol; j++) {
      disp_min = tab_disp_min[j];
      disp_max = tab_disp_max[j];

				/* (i + ni/2, j) in right image */
      pix_bas_d = q_im_d + i_S_ni2_P_nbcol_g + (j + disp_min);

      tab_produits_bas_ptr = tab_produits_bas + (j * delta_max_disp)
	+ (disp_min - disp_min_general);

      pixel = spix_1char(*pix_bas_g);
      for(d = disp_max - disp_min; d >= 0; d-=spix_vector_len(char)) {
	spix_correl_pixel_combine_vec(
	  tab_produits_bas_ptr, spix_ldu(char, pix_bas_d), pixel);
	tab_produits_bas_ptr += spix_vector_len(char);
	pix_bas_d += spix_vector_len(char);
      }

      pix_bas_g++;
    }

    /* the score for the first pixel of the line first */
    disp_min = tab_disp_nj2_min[j_min];
    disp_max = tab_disp_nj2_max[j_min];

    memset(somme_pdt_haut, 0, delta_max_disp * sizeof(SPIX_CORREL_SUM_TYPE));
    memset(somme_pdt_bas, 0, delta_max_disp * sizeof(SPIX_CORREL_SUM_TYPE));

    somme_pdt_bas_ptr = somme_pdt_bas + (disp_min - disp_min_general);
    somme_pdt_haut_ptr = somme_pdt_haut + (disp_min - disp_min_general);

    index = (j_min * delta_max_disp) + (disp_min - disp_min_general);
    tab_scores_ptr = tab_scores + index;
    tab_scores_2_ptr = tab_scores_2 + index;

    pix_somme_g = q_im_somme_g + j_min + i * nbcol;
    pix_somme_d = q_im_somme_d + (j_min +disp_min) + i * nbcol;

    pix_sigma_g = q_im_sigma_g + j_min + i * nbcol;
    pix_sigma_d = q_im_sigma_d + (j_min +disp_min) + i * nbcol;

    max_score = -1;
    for(d = disp_min; d <= disp_max; d++) {

      /* On se positionne en (d) dans les sommes des produits */
      tab_produits_bas_ptr = tab_produits_bas +  (d - disp_min_general);
      tab_produits_haut_ptr = tab_produits_haut + (d - disp_min_general);

      for (k = - nj2; k <= nj2; k++) {

	*somme_pdt_bas_ptr += *tab_produits_bas_ptr;
	*somme_pdt_haut_ptr += *tab_produits_haut_ptr;

	tab_produits_haut_ptr += delta_max_disp;
	tab_produits_bas_ptr += delta_max_disp;
      }

      *tab_scores_ptr +=  (*somme_pdt_bas_ptr) - (*somme_pdt_haut_ptr);

      *tab_scores_2_ptr =
	spix_correl_score(*tab_scores_ptr, prod_taille_inverse, *pix_somme_g,
			  *pix_somme_d, *pix_sigma_d, *pix_sigma_g);

      if (*tab_scores_2_ptr > max_score) {
	max_score = *tab_scores_2_ptr;
	tab_max[j_min] = d;
      }

      tab_scores_ptr++;
      tab_scores_2_ptr++;

      somme_pdt_haut_ptr++;
      somme_pdt_bas_ptr++;
      pix_somme_d++;
      pix_sigma_d++;
    }

    /* the remainging pixels of the line */

    for (j = j_min + 1; j < j_max; j++) {
      disp_min = tab_disp_nj2_min[j];
      disp_max = tab_disp_nj2_max[j];

      /* start with the minimum disparity, pixel (j, i) */
      pix_haut_g = q_im_g + ((i - ni2 - 1) * nbcol) + (j - nj2);
      pix_haut_d = q_im_d + ((i - ni2 - 1) * nbcol) + (j - nj2 + disp_min);

      pix_bas_g = q_im_g + ((i + ni2) * nbcol) + (j - nj2);
      pix_bas_d = q_im_d + ((i + ni2) * nbcol) + (j - nj2 + disp_min);

      somme_pdt_haut_ptr = somme_pdt_haut + (disp_min - disp_min_general);
      somme_pdt_bas_ptr = somme_pdt_bas + (disp_min - disp_min_general);

      somme_col_droite = 0;
      somme_col_gauche = 0;

      if (likely(nj < spix_vector_len(char))) {
	spix_vector(SPIX_CORREL_COMBINE_TYPE)
	  t[spix_vector_len(char)/spix_vector_len(SPIX_CORREL_COMBINE_TYPE)];
	spix_vector(SPIX_CORREL_COMBINE_TYPE)
	  b[spix_vector_len(char)/spix_vector_len(SPIX_CORREL_COMBINE_TYPE)];
	SPIX_CORREL_COMBINE_TYPE *tp = (SPIX_CORREL_COMBINE_TYPE *)t;
	SPIX_CORREL_COMBINE_TYPE *tb = (SPIX_CORREL_COMBINE_TYPE *)b;

	spix_correl_pixel_combine_vec(
	  tp, spix_ldu(char, pix_haut_g), spix_ldu(char, pix_haut_d));
	spix_correl_pixel_combine_vec(
	  tb, spix_ldu(char, pix_bas_g), spix_ldu(char, pix_bas_d));
	for(k = nj; k > 0; k--) {
	  somme_col_droite += *tp++;
	  somme_col_gauche += *tb++;
	}
      } else
	for(k = nj; k > 0; k--) {
	  somme_col_droite +=
	    spix_correl_pixel_combine(*(pix_haut_g++), *(pix_haut_d++));
	  somme_col_gauche +=
	    spix_correl_pixel_combine(*(pix_bas_g++), *(pix_bas_d++));
	}

      *somme_pdt_haut_ptr = somme_col_droite;
      *somme_pdt_bas_ptr = somme_col_gauche;


      pix_somme_g = q_im_somme_g + j + i * nbcol;
      pix_somme_d = q_im_somme_d + (j + disp_min) + i * nbcol;

      pix_sigma_g = q_im_sigma_g + j + i * nbcol;
      pix_sigma_d = q_im_sigma_d + (j + disp_min) + i * nbcol;

      index = (j * delta_max_disp) + (disp_min - disp_min_general);
      tab_scores_ptr = tab_scores + index;
      tab_scores_2_ptr = tab_scores_2 +  index;

      *tab_scores_ptr += (*somme_pdt_bas_ptr) - (*somme_pdt_haut_ptr) ;

      *tab_scores_2_ptr =
	spix_correl_score(*tab_scores_ptr, prod_taille_inverse, *pix_somme_g,
			  *pix_somme_d, *pix_sigma_d, *pix_sigma_g);

      max_score = *tab_scores_2_ptr;
      tab_max[j] = disp_min;

      /* compute score for the remaining disparities */
      pix_pdt_bas_gauche = tab_produits_bas
			   + ((j - nj2 - 1) * delta_max_disp)
			   + (disp_min + 1  - disp_min_general);
      pix_pdt_bas_droite = pix_pdt_bas_gauche + (nj * delta_max_disp);

      pix_pdt_haut_gauche = tab_produits_haut
			    + ((j - nj2 - 1) * delta_max_disp)
			    + (disp_min + 1  - disp_min_general);
      pix_pdt_haut_droite = pix_pdt_haut_gauche + (nj * delta_max_disp);

      index = (j * delta_max_disp) + (disp_min + 1 - disp_min_general);
      tab_scores_ptr = tab_scores + index;
      tab_scores_2_ptr = tab_scores_2 + index;

      somme_pdt_haut_ptr = somme_pdt_haut + (disp_min + 1 - disp_min_general);
      somme_pdt_bas_ptr = somme_pdt_bas + (disp_min + 1 - disp_min_general);

      pix_somme_d++;
      pix_sigma_d++;

      /* store these values as gcc isn't smart enough to guess they're
       * constant in the loops below */
      vec_suml = spix_1float(*pix_somme_g);
      vec_sigl = spix_1float(*pix_sigma_g);

      /* now compute scores for all disparities by using the previous
       * line. This is the deeper loop of the correlation and is executed
       * nblig x nbcol x delta_disp. */

      /* compute the first few unaligned scores */
      for(d = disp_min + 1;
	  !spix_isaligned(tab_scores_ptr) && d <= disp_max; d++) {
	SPIX_CORREL_SUM_TYPE s, s1, s2;

	s1 = *somme_pdt_bas_ptr +
	  *pix_pdt_bas_droite - *pix_pdt_bas_gauche;
	s2 = *somme_pdt_haut_ptr +
	  *pix_pdt_haut_droite - *pix_pdt_haut_gauche;
	s =  *tab_scores_ptr + s1 - s2;

	*tab_scores_2_ptr =
	  spix_correl_score(s, prod_taille_inverse, *pix_somme_g,
			    *pix_somme_d, *pix_sigma_d, *pix_sigma_g);
	if (*tab_scores_2_ptr > max_score) {
	  max_score = *tab_scores_2_ptr;
	  tab_max[j] = d;
	}

	*somme_pdt_bas_ptr = s1;
	*somme_pdt_haut_ptr = s2;
	*tab_scores_ptr= s;

	/* On se dcale d'une disparit dans le tableau des scores... */
	tab_scores_ptr++;
	tab_scores_2_ptr++;

	/* ... dans les tableaux des sommes de pdts... */
	somme_pdt_bas_ptr++;
	somme_pdt_haut_ptr++;

	pix_pdt_haut_gauche++;
	pix_pdt_haut_droite++;
	pix_pdt_bas_gauche++;
	pix_pdt_bas_droite++;

	/* and in right image */
       	pix_somme_d++;
       	pix_sigma_d++;
      }

      vec_max_score = spix_func_name(spix_1,SPIX_CORREL_SCORE_TYPE)(max_score);
      vec_max_ind   = spix_func_name(spix_1,SPIX_CORREL_SCORE_TYPE)(tab_max[j]);

      for(k=0; k<spix_vector_len(SPIX_CORREL_SCORE_TYPE); k++) {
	((SPIX_CORREL_SCORE_TYPE *)&vec_ind)[k] = d+k;
      }

      /* now the 16-bytes aligned disparities, with SSE if enabled */
      for(; d <= disp_max; d+=spix_vector_len(SPIX_CORREL_SUM_TYPE)) {
	spix_vector(SPIX_CORREL_SUM_TYPE) s, s1, s2;

	s1 = spix_lda(SPIX_CORREL_SUM_TYPE, somme_pdt_bas_ptr) +
	  spix_lda(SPIX_CORREL_SUM_TYPE, pix_pdt_bas_droite) -
	  spix_lda(SPIX_CORREL_SUM_TYPE, pix_pdt_bas_gauche);

	s2 = spix_lda(SPIX_CORREL_SUM_TYPE, somme_pdt_haut_ptr) +
	  spix_lda(SPIX_CORREL_SUM_TYPE, pix_pdt_haut_droite) -
	  spix_lda(SPIX_CORREL_SUM_TYPE, pix_pdt_haut_gauche);

	s = spix_lda(SPIX_CORREL_SUM_TYPE, tab_scores_ptr) + s1 - s2;

	spix_stra(tab_scores_2_ptr, SPIX_CORREL_SCORE_TYPE,
		  spix_correl_score_vec(s, prod_taille_inverse,
					vec_suml, spix_ldu(float, pix_somme_d),
					vec_sigl, spix_ldu(float, pix_sigma_d)));

	spix_func_name(spix_maxind_,SPIX_CORREL_SCORE_TYPE)
	  (spix_lda(SPIX_CORREL_SCORE_TYPE, tab_scores_2_ptr),
	   vec_ind, vec_max_score, vec_max_ind);

	spix_stra(tab_scores_ptr, SPIX_CORREL_SUM_TYPE, s);
	spix_stra(somme_pdt_haut_ptr, SPIX_CORREL_SUM_TYPE, s2);
	spix_stra(somme_pdt_bas_ptr, SPIX_CORREL_SUM_TYPE, s1);

	/* On se dcale d'une disparit dans le tableau des scores... */
	tab_scores_ptr+=	spix_vector_len(SPIX_CORREL_SUM_TYPE);
	tab_scores_2_ptr+=	spix_vector_len(SPIX_CORREL_SCORE_TYPE);

	/* ... dans les tableaux des sommes de pdts... */
	somme_pdt_bas_ptr+=	spix_vector_len(SPIX_CORREL_SUM_TYPE);
	somme_pdt_haut_ptr+=	spix_vector_len(SPIX_CORREL_SUM_TYPE);

	pix_pdt_haut_gauche+=	spix_vector_len(SPIX_CORREL_SUM_TYPE);
	pix_pdt_haut_droite+=	spix_vector_len(SPIX_CORREL_SUM_TYPE);
	pix_pdt_bas_gauche+=	spix_vector_len(SPIX_CORREL_SUM_TYPE);
	pix_pdt_bas_droite+=	spix_vector_len(SPIX_CORREL_SUM_TYPE);

	/* and in right image */
       	pix_somme_d+=		spix_vector_len(float);
       	pix_sigma_d+=		spix_vector_len(float);

	vec_ind+=
	  spix_func_name(spix_1,SPIX_CORREL_SCORE_TYPE)(
	    spix_vector_len(SPIX_CORREL_SUM_TYPE));
      }	/* Fin de la boucle sur les disparits */

      /* get the real max out of the max vector */
      for(k=0; k<spix_vector_len(SPIX_CORREL_SCORE_TYPE); k++) {
	if (((SPIX_CORREL_SCORE_TYPE *)&vec_max_score)[k] > max_score) {
	  max_score = ((SPIX_CORREL_SCORE_TYPE *)&vec_max_score)[k];
	  tab_max[j] = ((SPIX_CORREL_SCORE_TYPE *)&vec_max_ind)[k];
	}
      }
    } /* Fin de la boucle sur les colonnes */

    /* Recherche des disparits qui maximisent le score de corrlation pour tous
       les pixels de la ligne */
    spix_reversecheck(pix_disp, pix_sigma, pix_sigma_g_bis,
		    tab_max, tab_scores_2,
		    correlation_params, correlation_results,
		    nblig, j_min, j_max);

    /* done for the ith line */
    pix_disp++;
    if (pix_sigma) pix_sigma++;
    if (pix_sigma_g_bis) pix_sigma_g_bis+=	nbcol;

  } /* Fin de la boucle sur les lignes */

  return STEREO_OK;
}


/* --- sum_products_window ------------------------------------------- */

/** Sum the element of the array containing pixels products over the
 * [0:ni-1] x [j-nj/2:j+nj/2] window for the disparity d. */

static inline SPIX_CORREL_SUM_TYPE
sum_products_window(SPIX_CORREL_COMBINE_TYPE **plines, int j, int ni, int nj,
		    int d, int disp_min_general, int delta_max_disp)
{
  SPIX_CORREL_COMBINE_TYPE *pline;
  SPIX_CORREL_SUM_TYPE sum = 0;
  int k, l;

  for(l = ni-1; l >= 0; l--) {		/* loop over the lines */
    pline = plines[l] + (j-nj/2)*delta_max_disp + d - disp_min_general;
    for(k = nj; k > 0; k--) {	/* loop over the columns */
      sum += *pline;
      pline += delta_max_disp;
    }
  }

  return sum;
}


/* --- spix_reversecheck --------------------------------------------------- */

/** dtermine les disparits qui maximisent le score pour le tableau dtermin
 * pour toute une ligne de pixels. Avec vrification inverse.
 */

static void
spix_reversecheck(float *pix_disp, float *pix_sigma, float *pix_sigma_g,
		  int *tab_max, SPIX_CORREL_SCORE_TYPE *tab_scores,
		  SP_CORRELATION_PARAMS *correlation_params,
		  SP_CORRELATION_RESULTS *correlation_results,
		  int nblig, int j_min, int j_max)
{
  int      j, k;		/* Pour boucler sur les colonnes */

  int      disp_min, disp_max;	/* lments des tableaux ci-dessus */

  SPIX_CORREL_SCORE_TYPE *tab_scores_ptr;
  SPIX_CORREL_SCORE_TYPE *tab_scores_bis_ptr;

  float*   pix_disp_ptr;
  float*   pix_sigma_ptr;
  float*   pix_sigma_g_ptr;

  double   max_score;		/* Le max des scores */

  double   score_max_droite;
  double   score_max_gauche;	/* Les scores  droite et  gauche de
				   max_score */
  double   score_max_droite2;
  double   score_max_gauche2;	/* The correlation scores two pixels
				   before and after max_score */

  double   delta_voisins;	/* Sommes de diffrences de max_score avec ses
				 voisins */
  int      disparite_max;

  int    * tab_disp_max;
  int    * tab_disp_min;

  double   sigma;

  int      j_diag_h_d;
  int      d_diag_h_d;
  int      j_diag_b_g;

  int      jj; /* pour boucler dans le sens inverse */


  int delta_max_disp = correlation_results->delta_max_disp;
  int disp_min_general = correlation_results->disp_min_general;
  double seuil_sigma_inverse = 1.0 / correlation_params->var_th;
  double seuil_score = correlation_params->score_th;
  double seuil_voisins = correlation_params->diff_neighbor_max_th;
  int epsilon_inverse = correlation_params->inv_th;

  /* On se positionne  la colonne j_min+1 dans l'image de disparit et l'image
     de sigmas */
  pix_disp_ptr = pix_disp + nblig;
  pix_sigma_ptr = pix_sigma + nblig;
  pix_sigma_g_ptr = pix_sigma_g + 1;

  tab_scores_bis_ptr
    = tab_scores + ((j_min+1) * delta_max_disp) - disp_min_general;

  tab_disp_min = &correlation_results->tab_disp_nj2_min[j_min+1];
  tab_disp_max = &correlation_results->tab_disp_nj2_max[j_min+1];

  for (j = j_min + 1; j < j_max -1; j++,
	 tab_disp_max++,
	 tab_disp_min++,
	 pix_disp_ptr += nblig,
	 pix_sigma_ptr += nblig,
	 pix_sigma_g_ptr++,
	 tab_scores_bis_ptr += delta_max_disp) {

    if (pix_sigma_g && *pix_sigma_g_ptr > seuil_sigma_inverse)
      continue;

    disp_min = *tab_disp_min;
    disp_max = *tab_disp_max;

    /* get the maximum score and the best disparity */
    disparite_max = tab_max[j];
    if (disparite_max >= disp_max || disparite_max <= disp_min) continue;

    /* apply thresholds on the max */
    max_score = tab_scores_bis_ptr[disparite_max];
    if (max_score < seuil_score) continue;

    score_max_gauche = tab_scores_bis_ptr[disparite_max - 1];
    score_max_droite = tab_scores_bis_ptr[disparite_max + 1];

    delta_voisins = 2.*max_score - score_max_gauche - score_max_droite;
    if (delta_voisins < 2. * seuil_voisins) continue;

#if 0
    /* reverse correlation check: check that we have a local
     * maximum around the max disparity */
    if (max_score - seuil_voisins <=
	tab_scores_bis_ptr[disparite_max + 1 - delta_max_disp] ||
	max_score - seuil_voisins <=
	tab_scores_bis_ptr[disparite_max - 1 + delta_max_disp]) {

      /* or near the max disparity (up to epsilon_inverse) */
      for(k=1; k <= epsilon_inverse; k++) {
	if (tab_scores_bis_ptr[disparite_max + k*(1-delta_max_disp)] - seuil_voisins >
	    tab_scores_bis_ptr[disparite_max + (k+1)*(1-delta_max_disp)] &&
	    tab_scores_bis_ptr[disparite_max + k*(1-delta_max_disp)] - seuil_voisins >
	    tab_scores_bis_ptr[disparite_max + (k-1)*(1-delta_max_disp)])
	  break;
	if (tab_scores_bis_ptr[disparite_max + k*(delta_max_disp-1)] - seuil_voisins >
	    tab_scores_bis_ptr[disparite_max + (k+1)*(delta_max_disp-1)] &&
	    tab_scores_bis_ptr[disparite_max + k*(delta_max_disp-1)] - seuil_voisins >
	    tab_scores_bis_ptr[disparite_max + (k-1)*(delta_max_disp-1)])
	  break;
      }
      if (k > epsilon_inverse) {
	correlation_results->nbRemovedByUnicity++;
	continue;
      }
    }
#endif

#if 1
    /* do the full reverse correlation check */

    /* get the upper right diagonal limit */
    if ( (j + disparite_max - disp_min_general) > j_max ) {
      /* La diagonale intersecte la verticale droite du tableau */
      j_diag_h_d = j_max;
      d_diag_h_d = j + disparite_max - disp_min_general - j_max;
    } else {
      /* La diagonale intersecte l'horizontale haute du tableau */
      j_diag_h_d = j + disparite_max - disp_min_general;
      d_diag_h_d = 0;
    }

    tab_scores_ptr = tab_scores_bis_ptr + disparite_max + 2*(delta_max_disp-1);

    for (jj = j_diag_h_d - j + 1; jj > 0; jj--,
	   tab_scores_ptr += delta_max_disp-1)
      if (*tab_scores_ptr >= max_score) break;
    if (jj > epsilon_inverse) continue;

    /* get the lower left diagonal limit */
    if ( (j + disparite_max - disp_min_general - delta_max_disp) < j_min) {
      /* La diagonale intersecte la verticale gauche du tableau */
      j_diag_b_g = j_min;
    } else {
      /* La diagonale intersecte l'horizontale basse du tableau */
      j_diag_b_g = j + disparite_max - disp_min_general - delta_max_disp;
    }

    tab_scores_ptr = tab_scores_bis_ptr + disparite_max + 2*(1-delta_max_disp);

    for (jj = j - 1 - j_diag_b_g; jj > 0; jj--,
	   tab_scores_ptr += 1-delta_max_disp)
      if (*tab_scores_ptr >= max_score) break;
    if (jj > epsilon_inverse) continue;
#endif

    /* On interpole la disparit */
    switch (correlation_params->type_interpolation) {
      case SP_INTERPOLATION_DISP_NIL:
	/* Just put the integer disparity */
	*(pix_disp_ptr) = disparite_max;
	break;

      case SP_INTERPOLATION_DISP_PARABOLIQUE_ONLY:
	/* Do the parabolic interpolation */
	*(pix_disp_ptr) = disparite_max
	  + ((score_max_droite - score_max_gauche)
	     / (2.0 * delta_voisins));
	break;

      case SP_INTERPOLATION_DISP_BILINERAIRE_ONLY:
	/* Do the "bi-linear" interpolation if possible */
	if ( ((disparite_max - disp_min) >= 2)
	     && ((disp_max - disparite_max) >= 2) ) {
	  /* Yes, it is possible to perform the bilinear interpolation */
	  double pipo;
	  double denominator;
	  score_max_gauche2 = *(tab_scores_bis_ptr + disparite_max - 2);
	  score_max_droite2 = *(tab_scores_bis_ptr + disparite_max + 2);
	  denominator = ((score_max_gauche - score_max_gauche2) +
			 (score_max_droite - score_max_droite2));
	  if (denominator != 0) {
	    pipo = (2.0 * (score_max_droite - score_max_gauche)
		    + score_max_gauche2 - score_max_droite2) /
	      ((score_max_gauche - score_max_gauche2) +
	       (score_max_droite - score_max_droite2));
	    /* Do the bilinear interpolation */
	    *(pix_disp_ptr) = disparite_max + pipo;
	  }
	}
	break;

      case SP_INTERPOLATION_DISP_BOTH:
	/* The bilinear interpolation is checked, and applied if
	   it is applicable. If not, (i.e. if the correction
	   exceeds 1 or the "2nd order" neighbors have not be
	   computed) then the parabolic interpolation is applied */
	if ( ((disparite_max - disp_min) >= 2)
	     && ((disp_max - disparite_max) >= 2) ) {
	  double pipo;
	  double denominator;
	  score_max_gauche2 = *(tab_scores_bis_ptr + disparite_max - 2);
	  score_max_droite2 = *(tab_scores_bis_ptr + disparite_max + 2);
	  denominator = ((score_max_gauche - score_max_gauche2) +
			 (score_max_droite - score_max_droite2));
	  if (denominator != 0) {
	    pipo = (2.0 * (score_max_droite - score_max_gauche)
		    + score_max_gauche2 - score_max_droite2) /
	      denominator;
	    if ((pipo >= 1.0) || (pipo <= -1.0)) {
	      /* The correction exceeds 1: we do the parabolic interpolation */
	      *(pix_disp_ptr) = disparite_max
		+ ((score_max_droite - score_max_gauche)
		   / (2.0 * delta_voisins));
	    } else
	      /* The bi-linear interpolation is done */
	      *(pix_disp_ptr) = disparite_max + pipo;
	  } else
	    /* The denominator equals 0: : we do the parabolic interpolation */
	    *(pix_disp_ptr) = disparite_max
	      + ((score_max_droite - score_max_gauche)
		 / (2.0 * delta_voisins));
	} else {
	  /* Both 2nd order neighbors do not exist: we do the parabolic interpolation */
	  *(pix_disp_ptr) = disparite_max
	    + ((score_max_droite - score_max_gauche)
	       / (2.0 * delta_voisins));
	}
	break;
    }

    if (pix_sigma) {
      /* Estimation (avec le modle empirique) de l'cart type sur la
	 disparit */
      sigma = MIN(max_score - score_max_gauche, max_score - score_max_droite);
      /* TODO : passer les paramtres du modle dans les paramtres de
	 la fonction */
      *(pix_sigma_ptr) = (0.3426 * sigma * sigma) - (0.2405 * sigma) + 0.0554;
    }
  }
}
