/*
 * Copyright (c) 2000,2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                                    --  Fri Aug  4 2006
 */

/* librairies standards */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Les erreurs, structures et numrations locales */
#include "libstereopixel.h"

/* Les protos et erreurs d'autres libs. */
#include "libimages3d.h"

/* Flags de compilations */

#define OPTIMISATION  2    /* 0: pas d'optimisation, 1: medium, 2: best */
#define VERBOSE       0    /* 0: silencieux, puis 1 a 3 */
#define RE_ETIQUETAGE_OPTIONNEL 1  /* 0 ou 1 */


#define NB_MAX_REGIONS 100000	/* N'ayons pas peur */

/*************************************************************
  D'aprs la fonction extractPoints de binetique.c, revue par Sara en Mars
  1998, lgrement radapte par Simon en Mars 98 pour interface calife, puis
  optimise par Sara.

  Etiquetage en composantes connexes des pixels de valeur proche (a diffMax
  prs) NON NULLE (pixels nuls considrrs comme non significatifs) d'une
  image de float.

  Les composants ayants moins de sizeMin pixels sont mis a zero.

  parametres de la fonction
  entre :
  - imf       image de float calife (range par colonnes!)

  - im_region : l'image des rgions dtermines

  - images_size : la structure o tout est dit concernant les tailles des l'image

  - sizeMin   taille mini des regions

  - diffMax   ecart max entre les valeurs de deux pixels d'une mme rgion

  sortie :

  - ERR_OK ou ERR_SP_TOO_MANY_BLOBS

  ***************************************************************/

static void char_miseajour_equiva(int oldval, int newval,
				  int *tabequiv, int max_value);

int blob_filter_disparites (SP_CORRELATION_RESULTS*	correlation_results,
				  int			sizeMin,
				  float			diffMax)

{
  int   nb_lig;
  int   nb_col;			/* Taille de l'image */

  float*	im_float;
  int*		im_region;	/* Les images rcupres dans la structure
				   'correlation_results' */

  int   i_min;
  int   i_max;
  int   j_min;
  int   j_max;			/* Bornes pour boucler */

  int   i, j;			/* Indices pour boucler */

  float*   pixel_val_ptr;
  float    pixel_val;
  float    pixel_val_haut;
  float    pixel_val_gauche;	/* Ptr sur le pixel courant et valeurs de ses
				   voisins dans l'image  filtrer*/
  float pixel_val_diff;

  int*     pixel_coul_ptr;
  int      pixel_coul_haut;
  int      pixel_coul_gauche;	/* Ptr sur le pixel courant et valeur de ses
				   voisins dans l'image de rgions */

  int couleur =  0;
  int maxcoul =  0;
  int newcoul = -1;

  /* Tableau couleur equivalente */
  static int tabequiv[NB_MAX_REGIONS];

  /* Nombre de points par couleur */
  static int  n[NB_MAX_REGIONS];

  /* Flag couleur utilise pour les equivalences */
  static char used[NB_MAX_REGIONS];

#if VERBOSE
  int nb_blobs = 0; /* Quelques infos pour le mode verbose */
#endif /* VERBOSE */


  /* On rcupre les tailles des images... */
  nb_lig = correlation_results->images_sizes.nb_lig_final;
  nb_col = correlation_results->images_sizes.nb_col_final;

  /* ... et les images elles-mmes. */
  im_float = correlation_results->im_disparites;
  im_region = correlation_results->im_regions;


  i_max = nb_lig - 1;
  j_max = nb_col - 1;
  i_min = 1;
  j_min = 1;

  /* On initialise  0 l'image des rgions */
  bzero((char*) im_region, nb_lig * nb_col * sizeof(int));
  bzero((char*) tabequiv,  NB_MAX_REGIONS  * sizeof(int));
  bzero((char*) n,         NB_MAX_REGIONS  * sizeof(int));
  bzero((char*) used,      NB_MAX_REGIONS  * sizeof(char));

  /* Reset nbCorrelated point */
  correlation_results->nbCorrelated = 0;

  /* On se positionne en (1,1), et on boucle. Attention, c'est une boucle pour
     des images ranges comme des ima2 */
  pixel_val_ptr = im_float - 2 + nb_lig;
  pixel_coul_ptr = im_region - 2 + nb_lig;

  for(j = j_min; j < j_max; j++) {

    pixel_val_ptr += 2;
    pixel_coul_ptr += 2;

    for(i = i_min; i < i_max; i++) {

      pixel_val_ptr++;
      pixel_coul_ptr++;

      pixel_val = *pixel_val_ptr;

      /* Pas correle: on vire */
      if(pixel_val == 0.0)
	continue;

      pixel_coul_gauche = *(pixel_coul_ptr - nb_lig);

      /* Appartenance au blob de gauche ? */
      if (pixel_coul_gauche != 0) {

	pixel_val_gauche = *(pixel_val_ptr - nb_lig);
	pixel_val_diff = pixel_val - pixel_val_gauche;
 	if ( (pixel_val_diff < diffMax) && (- diffMax < pixel_val_diff)) {

	  /* On prend la couleur du blob gauche */
	  couleur = tabequiv [pixel_coul_gauche];

	  /* Fusion des blobs de gauche et du haut ? */
	  pixel_coul_haut = *(pixel_coul_ptr - 1);
	  if (pixel_coul_haut != 0 && tabequiv[pixel_coul_haut] != couleur) {
	    pixel_val_haut = *(pixel_val_ptr - 1);
	    pixel_val_diff = pixel_val - pixel_val_haut;
	    if ( (pixel_val_diff < diffMax) && (-diffMax < pixel_val_diff)) {

	      /* Fusion */
#if VERBOSE
	      nb_blobs--;
#endif

#if OPTIMISATION
	      /* blob pixel_coul_haut pas utilis par ailleurs: on le met
		 directement a couleur */
	      if (!used[tabequiv[pixel_coul_haut]]) {

#if       OPTIMISATION==2
		/* Un seul pixel pour ce blob: on le met directement a coul */
		if (n[pixel_coul_haut] == 1) {
		  *(pixel_coul_ptr - 1) = couleur;
		  n[couleur]++;
		  n[pixel_coul_haut] = 0;
		  tabequiv[pixel_coul_haut] = 0;
		  newcoul = pixel_coul_haut;
		}

		/* Fusion de ce blob avec coul */
		else
#endif  /* OPTIMISATION==2 */
		  {
		  tabequiv[pixel_coul_haut] = couleur;
		  used[couleur] = 1;
		}
	      }
	      else {
		/* pixel_coul_gauche pas utilis par ailleurs: on le met
		   directement a pixel_coul_haut */
		if (!used[couleur]) {

#if      OPTIMISATION==2
		/* Un seul pixel pour ce blob: on le met directement a coul */
		if (n[couleur] == 1) {
		  *(pixel_coul_ptr - nb_lig) = tabequiv[pixel_coul_haut];
		  n[tabequiv[pixel_coul_haut]]++;
		  n[couleur] = 0;
		  tabequiv[couleur] = 0;
		  newcoul = couleur;
		  couleur = tabequiv[pixel_coul_haut];
		}
		/* Fusion de ce blob avec coul */
		else
#endif  /* OPTIMISATION==2 */
		  {
		  tabequiv[pixel_coul_gauche] = tabequiv[pixel_coul_haut];
		  couleur = tabequiv[pixel_coul_gauche];
		}

		}
		/* Les deux couleurs servent: on les fusionne */
		else {
		  used[tabequiv[pixel_coul_haut]] = 0;
#endif /* OPTIMISATION */
		  char_miseajour_equiva(tabequiv[pixel_coul_haut], couleur,
					tabequiv, maxcoul);
#if OPTIMISATION
		}
	      }
#endif /* OPTIMISATION */
	    }
	  } /* Fin fusion */

	  /* Enregistre la couleur */
	  *pixel_coul_ptr = couleur;
	  n[couleur]++;
	  continue;
	}
      }

      /* Appartenance au blob du haut ? */
      pixel_coul_haut = *(pixel_coul_ptr - 1);
      if (pixel_coul_haut != 0) {
	pixel_val_haut = *(pixel_val_ptr - 1);
	pixel_val_diff = pixel_val - pixel_val_haut;
	if ((pixel_val_diff < diffMax) && (-diffMax < pixel_val_diff)) {
	  couleur = tabequiv[pixel_coul_haut];
	  *pixel_coul_ptr = couleur;
	  n[couleur]++;
	  continue;
	}
      }

      /* Nouveau blob */
#if       OPTIMISATION==2
      if (newcoul != -1) {
	couleur = newcoul;
	newcoul  = -1;
      }
      else
#endif  /* OPTIMISATION==2  */
	{
	maxcoul++;
	if (maxcoul == NB_MAX_REGIONS) {
	  return stereoError(ERR_SP_TOO_MANY_BLOBS);
	}
	couleur = maxcoul;
      }
#if VERBOSE
      nb_blobs++;
#endif
      tabequiv[couleur] = couleur;
      *pixel_coul_ptr = couleur;
      n[couleur]++;

    } /* Fin du second for */
  } /* Fin du premier for */

  /*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
  /* Fusion des regions equivalentes  */
  /*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
  for(couleur = 1; couleur <= maxcoul ; couleur++)
    if(couleur != tabequiv[couleur])
      n[tabequiv[couleur]] += n[couleur];


  /*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
  /* Affiche la liste des rgions */
  /*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
#if VERBOSE>1
  {
    int nb_pix_removed_ctrl = 0;
    int nb_blobs_removed = 0;
    for(couleur = 1; couleur <= maxcoul ; couleur++) {
      if(couleur == tabequiv[couleur]) {
#if VERBOSE!=3
	if (n[couleur] > 5)
#endif /* VERBOSE!=3 */
	  printf("couleur %5d    nbp %6d", couleur, n[couleur]);
	if (n[couleur] < sizeMin) {
	  nb_blobs_removed++;
	  nb_pix_removed_ctrl +=  n[couleur];
#if VERBOSE!=3
	  if (n[couleur] > 5)
#endif /* VERBOSE!=3 */
	    printf("\n");
	}
	else
	  printf("  (conserve)\n");
      }
    }
    printf ("Total: %d suppressed regions (%d pixels)\n\n",
	       nb_blobs_removed, nb_pix_removed_ctrl);
  }
#endif /* VERBOSE>1 */


  /*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
  /* Supression des petites regions */
  /*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
  /* On se positionne en (1,1) dans les images */
  pixel_coul_ptr = im_region + 1 + nb_lig;
  pixel_val_ptr = im_float + 1 + nb_lig;
  for(j = j_min; j < j_max; j++) {
    for(i = i_min; i < i_max; i++){
      couleur = *pixel_coul_ptr;
      if(couleur != 0 && n[tabequiv[couleur]] < sizeMin) {
	*pixel_val_ptr = 0.0;
	correlation_results->nbRemovedByBlob++;
      }
      else if(couleur != 0 && n[tabequiv[couleur]] >= sizeMin) {
	correlation_results->nbCorrelated++;
      }
      pixel_coul_ptr++;
      pixel_val_ptr++;
    }
    pixel_coul_ptr += 2;
    pixel_val_ptr += 2;
  }


  /*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
  /* Affichage bilan   */
  /*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
#if VERBOSE
  printf ("%5d rgions au total   (%d couleurs attribues)\n"
	  "%5d pixels supprims sur %d (%f pourcent)\n",
	  nb_blobs, maxcoul,
	  correlation_results->nbRemovedByBlob, nb_col*nb_lig,
	  (double)correlation_results->nbRemovedByBlob*100./(double)(nb_col*nb_lig));
#endif /* VERBOSE */


  /*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
  /* Re-etiquetage optionel pour visualiser les regions */
  /*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
#if RE_ETIQUETAGE_OPTIONNEL
  /* On se positionne en (1,1) dans les images */
  pixel_coul_ptr = im_region - 2 + nb_lig;
  for(j = j_min; j < j_max; j++) {
    pixel_coul_ptr += 2;
    for(i = i_min; i < i_max; i++) {
      pixel_coul_ptr++;
      couleur = *pixel_coul_ptr;
      if (couleur == 0)
	continue;
      if (n[tabequiv[couleur]] >= sizeMin)
	*pixel_coul_ptr = 0.0;
      else if(couleur != tabequiv[couleur])
	*pixel_coul_ptr = tabequiv[couleur];
    }
  }
#endif /* RE_ETIQUETAGE_OPTIONNEL */

  /* Mise  jour de l'tat de l'image des rgions */
  correlation_results->state_regions = SP_IMAGE_COMPUTED;

  return (STEREO_OK);
}

/**************************************************************

  char_miseajour_equiva :

  fonction mise a jour de la table d'equivalence des regions

  ***************************************************************/
static void char_miseajour_equiva(int oldval, int newval,
				  int *tabequiv, int max_value)
{
  int ind;
  int *tmp = tabequiv+1;
  for(ind = 1; ind <= max_value; ind++, tmp++)
    if(*tmp == oldval)
      *tmp = newval;
}
