

/*
 * Copyright (c) 2000,2011 LAAS/CNRS                       --  Fri Aug  4 2006
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

/* librairies standards */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Les erreurs, structures et numrations locales */
#include "libstereopixel.h"

/* Les protos et erreurs d'autres libs. */
#include "libimages3d.h"

/* La config de libstereopixel */
#include "../stereopixel-config.h"


/****************************************************************************
 *
 * Fonction qui cree une image 3D (type DATA_IM3D)  partir de l'image de *
 * disparits, et l'exprime dans le repre camra, ou camra rectifie, et
 * calcule ventuellement une matrice de variance pour chacun des points (tout
 * a suivant les options 'option_variances' et 'option_repere')
 *
 ****************************************************************************/

int
reconstruction_3d (SP_CORRELATION_RESULTS*	correlation_results,
		   SP_IMAGES_SIZES*		images_size,
		   SP_REPERE_RECONSTRUCTION	type_repere,
		   SP_TYPE_RECONSTRUCTION	type_recontruction,
		   double*			rect_2_cam,
		   double*			baseline,
		   double*			mat_intrinseque,
		   double*			mat_intrinseque_rectifiee,
		   DATA_IM3D*			im3d)

{
  int           nblig;
  int           nbcol;
  int           reduction_i;
  int           reduction_j;
  int		i_min;
  int		j_min;
  int		i_max;
  int		j_max;		/* Tailles de l'image 3D, extraites de
				   'images_size' */

/*   double        transfo[16]; */

  float*	q_im_disp;
  float*	q_im_sigma;	/* Retrouvs dans 'correlation_results' */
  float*        pix_disp;
  float*        pix_sigma;	/* Un pixel de l'image de disparits (et des
				   sigmas) */

  DATA_PT3D*    point;		/* Un point de l'image 3D */


  /* Parametres de calibration intrinsque de l'image rectifie */
  double        u0, v0;
  double        ku, kv;
  /* Baseline */
  double        base;

  double        tr0, tr1, tr2, tr3, tr4, tr5, tr6, tr7, tr8;
				/* Les 9 lments de 'rect_2_cam' qui
				   serviront  ``drectifier'' l'image 3D (on
				   les introduit pour optimiser un peu) */

  int		i,j;		/* Pour boucler sur l'image de disparits */
  double	x;
  double	y;
  double	z;		/* Coordonnes d'un pt3d (dans le repre
				   'camera_rectifiee'*/

  double	alpha;
  double	beta;
  double	gamma;
  double	facteur;	/* Pour le calcul des covariances. (x =
				   alpha/disp, y = beta*x, et z =
				   gamma*x). Par ailleurs, tous les lments
				   de la matrice peuvent tre mis en facteur
				   (par la variable 'facteur') */

  /* Suivent des variables d'optimisation, pour ne pas les recalculer dans la
     boucle */
  double	base_reduite;
  int		i_min_opt;
  int		i_max_opt;
  int		j_min_opt;
  int		j_max_opt;	/* Bornes de bouclage, attention, elles
				   comprennent sont multiplie par les taux de
				   rduction */

  /* Validit des paramtres */
  /* Les point sont-ils bien allous ? */
  if (im3d->header.points == NIL_POINTS)
    return stereoError(ERR_SP_ALLOC_IM3D);
  /* L'image de disparits est-elle calcule ? */
  if (correlation_results->state_disparites != SP_IMAGE_COMPUTED)
    return stereoError (ERR_SP_CORRELATION_RESULTS_NOT_COMPUTED);
  /* Les carts types sur la disparit sont dtermins ? */
  if ((type_recontruction == SP_RECONSTRUCTION_SIGMA)
      && (correlation_results->state_sigma_disparites != SP_IMAGE_COMPUTED))
    return stereoError (ERR_SP_CORRELATION_RESULTS_NOT_COMPUTED);
  /* Les matrices de covariances sont-elles alloues ? */
  if ((type_recontruction == SP_RECONSTRUCTION_SIGMA)
      && (im3d->header.covariances == NIL_COVARIANCES))
    return stereoError(ERR_SP_ALLOC_IM3D);

  /* On retrouve les tailles de l'image */
  i_min = images_size->i_min;
  j_min = images_size->j_min;
  i_max = images_size->i_max;
  j_max = images_size->j_max;
  nblig = images_size->nb_lig_final;
  nbcol = images_size->nb_col_final;
  reduction_i = images_size->taux_reduction_i;
  reduction_j = images_size->taux_reduction_j;

  q_im_disp = correlation_results->im_disparites;
  q_im_sigma = correlation_results->im_sigma_disparites;

  /* Mise  jour des champs de l'entte de l'image */
  im3d->header.capteur = CAPTEUR_STEREO;
  im3d->header.repere = REPERE_CAPTEUR;
  im3d->header.coordonnees = COORDONNEES_CARTESIENNES;
  if ((reduction_i != 1) && (reduction_j != 1))
    /* Il y a eu rduction */
    im3d->header.reduction = REDUCTION;
  /* TODO : mettre un flag 'sous_images' dans les im3d ? */
  else
    im3d->header.reduction = NIL_REDUCTION;

  /* TODO XXX !!! : mettre un second champ de reduction dans les images 3D ? */

  if ((i_min == 0)
      && (images_size->i_max == images_size->nb_lig_final)
      && (j_min == 0)
      && (images_size->j_max == images_size->nb_col_final))
    /* Il n'y a pas eu d'extraction de sous-image */
    im3d->header.sous_image = NIL_SOUS_IMAGE;
  else
    im3d->header.sous_image = SOUS_IMAGE;

  /* Mise  jour des tailles de l'image */
  im3d->header.nblig_initial = images_size->nb_lig_original;
  im3d->header.nbcol_initial = images_size->nb_col_original;
  im3d->header.nblig = images_size->nb_lig_final;
  im3d->header.nbcol = images_size->nb_col_final;
  /* Attention  la signification spciale des min et max dans les images 3D :
   on multiplie les valeurs correspondantes des images stro par les taux de
   rduction. TODO XXX !!! : vrifier (et homogniser ?) tout a ! */
  im3d->header.i_min = reduction_i * images_size->i_min;
  im3d->header.j_min = reduction_j * images_size->j_min;
  im3d->header.i_max = reduction_i * images_size->i_max;
  im3d->header.j_max = reduction_j * images_size->j_max;

  im3d->header.taux_reduction = reduction_i;

  /* Recuperation des parametres calculs a priori au moment du calcul des
     matrices de rectifications */
  u0 = mat_intrinseque_rectifiee[2] / images_size->previousReduction;
  v0 = mat_intrinseque_rectifiee[5] / images_size->previousReduction;
  ku = mat_intrinseque_rectifiee[0] / images_size->previousReduction;
  kv = mat_intrinseque_rectifiee[4] / images_size->previousReduction;
  base = baseline[1]*ku;	/* = baseline . scale j */

  tr0 = rect_2_cam[0];
  tr1 = rect_2_cam[1];
  tr2 = rect_2_cam[2];
  tr3 = rect_2_cam[3];
  tr4 = rect_2_cam[4];
  tr5 = rect_2_cam[5];
  tr6 = rect_2_cam[6];
  tr7 = rect_2_cam[7];
  tr8 = rect_2_cam[8];

  /* Let's go. On fait des boucles diffrentes suivant la valeur des flags
     'type_recontruction' et 'type_repere' juste pour ne pas  tester leur
     valeur  chaque point. a allonge un peu le code, mais a gagne du
     temps  l'excution... TODO : cette optimisation m'a pos bcp de pbs !!!
     Finalement, je ne l'ai pas faite */
  i_min_opt = (reduction_i * images_size->i_min);
  i_max_opt = (reduction_i * (images_size->i_max - 1));
  j_min_opt = (reduction_j * images_size->j_min);
  j_max_opt = (reduction_j * (images_size->j_max - 1));

  /* Dtermination des variables d'optimisation */
  base_reduite = base / (reduction_i * images_size->previousReduction);

  if (type_repere == SP_REPERE_CAMERA) {
    /* Les points sont exprims dans le repre de la camra originale */
    pix_disp = q_im_disp;
    for (j = 0; j < nbcol; j++) {
      point = im3d->points_tab + j;
      for (i = 0; i < nblig; i++) {

	if (*pix_disp != 0.0) {
	  point->state = GOOD_PT3D;
	  point->coordonnees = COORDONNEES_CARTESIENNES;
	  point->repere = REPERE_CAPTEUR;

	  /* Attention : la matrice de rotation est en fait bizarement foutue
	     (atavisme li au temps o le paramtre Z codait la profondeur des
	     points), d'o le calcul trange qui suit ( comparer avec le cas
	     o les donnes sont exprimes dans le repre rectifi, voir plus
	     bas). TODO : il serait bien de mettre a au propre, en corrigeant
	     la manire dont est calcule la matrice dans la fonction
	     'estim_rectification_params' */

	  z = - base_reduite / (*pix_disp);
	  x = z * ((reduction_j * (j + j_min)) - u0) / ku;
	  y = z * ((reduction_i * (i + i_min)) - v0) / kv;

	  /* ...et on applique la rotation. */
	  point->coord_1 = (tr0 * x) + (tr3 * y) + (tr6 * z);
	  point->coord_2 = (tr1 * x) + (tr4 * y) + (tr7 * z);
	  point->coord_3 = (tr2 * x) + (tr5 * y) + (tr8 * z);
	}
	else
	  /* Disparit invalide */
	  point->state = BAD_PT3D;

	pix_disp++;
	point += nbcol;

      }
    } /* Fin de la boucle */
  }
  else {
      /* Les points sont exprims dans le repre de la camra
	 rectifie. Remarque : ici, le corps de la boucle a t optimis */
    if (type_recontruction == SP_RECONSTRUCTION_SIGMA) {
      /* On rempli la matrice de covariances. */
      pix_disp = q_im_disp;
      pix_sigma = q_im_sigma;
/*       point_bis = im3d->points_tab; */
      alpha = base_reduite;
      for (j = 0; j < nbcol; j++) {
	point = im3d->points_tab + j;
	for (i = 0; i < nblig; i++) {

	  if (*pix_disp != 0.0) {
	    point->state = GOOD_PT3D;
	    point->coordonnees = COORDONNEES_CARTESIENNES;
	    point->repere = REPERE_CAPTEUR;
	    point->covariances = COVARIANCES_CALCULEES;

	    x = - alpha / (*pix_disp);

	    beta = ((reduction_j * (j + j_min)) - u0) / ku;
	    gamma = ((reduction_i * (i + i_min) - v0)) / kv;

	    point->coord_1 = x * beta;
	    point->coord_2 = x * gamma;
	    point->coord_3 = x;

	    /* Remplissage de la matrice de covariance */
	    facteur = - (*pix_sigma) * x / (*pix_disp);
	    facteur *= facteur;

	    point->covariances_tab[0] = facteur * beta * beta;
	    point->covariances_tab[1] = facteur * beta * gamma;
	    point->covariances_tab[2] = facteur * beta;
	    point->covariances_tab[3] = facteur * gamma * gamma;
	    point->covariances_tab[4] = facteur * gamma;
	    point->covariances_tab[5] = facteur;
	  }
	  else
	    /* Disparit invalide */
	    point->state = BAD_PT3D;

	  pix_disp++;
	  pix_sigma++;
	  point += nbcol;
	}
/* 	point_bis++; */
      } /* Fin de la boucle */
    }
    else {
      /* On ne rempli pas la matrice de covariances. */
      pix_disp = q_im_disp;
      for (j = 0; j < nbcol; j++) {
	point = im3d->points_tab + j;
	for (i = 0; i < nblig; i++) {

	  if (*pix_disp != 0.0) {
	    point->state = GOOD_PT3D;
	    point->coordonnees = COORDONNEES_CARTESIENNES;
	    point->repere = REPERE_CAPTEUR;

	    x = - base_reduite / (*pix_disp); /* Signe '-' car on a des
						 disparits ngatives */

	    point->coord_1 = x * ((reduction_j * (j + j_min) - u0)) / ku;
	    point->coord_2 = x * ((reduction_i * (i + i_min) - v0)) / kv;
	    point->coord_3 = x;
	  }
	  else
	    /* Disparit invalide */
	    point->state = BAD_PT3D;

	  pix_disp++;
	  point += nbcol;
	}
      } /* Fin de la boucle */
    }
  }

  /* Copie de la matrice de calibration intrinsque */
  for (i = 0; i < 16; i++)
    im3d->header.mat_intrinseque[i] = *(mat_intrinseque_rectifiee + i);

  im3d->header.points = POINTS_CALCULES;
  if (type_recontruction == SP_RECONSTRUCTION_SIGMA)
    im3d->header.covariances = COVARIANCES_CALCULEES;


  return (STEREO_OK);
}

