/*	$LAAS$ */

/*
 * Copyright (c) 2000,2008 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                                        Fri Aug  4 2006
 */

#include "stereopixel-config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libstereopixel.h"
#if SPIX_THREADS
# include "workers.h"
#endif

/***********************************************************************
 *
 * Fonction de cration d'une structure 'SP_CORRELATION_PARAMS'
 *
 ***********************************************************************/

SP_CORRELATION_PARAMS*
create_correlation_params () {
  SP_CORRELATION_PARAMS* new_correlation_params;

  new_correlation_params
    = (SP_CORRELATION_PARAMS*) malloc (sizeof (SP_CORRELATION_PARAMS));

  if (new_correlation_params == NULL) {
    stereoError (ERR_SP_MALLOC_CORRELATION_PARAMS);
    return (NULL);
  }

  new_correlation_params->state = SP_CORRELATION_PARAMS_NOT_INITIALIZED;

  return (new_correlation_params);
}

/***********************************************************************
 *
 * Fonction de destruction d'une structure 'SP_CORRELATION_PARAMS'
 *
 ***********************************************************************/

void
destroy_correlation_params (SP_CORRELATION_PARAMS* correlation_params)
{
  if (correlation_params == NULL)
    return;

#if SPIX_THREADS
  if (correlation_params->state == SP_CORRELATION_PARAMS_INITIALIZED)
    if (correlation_params->threads > 1 &&
	correlation_params->thread_array) {
      spix_worker *w;
      int i, t;

      w = correlation_params->thread_array[0];
      pthread_barrier_destroy(w->start);
      pthread_barrier_destroy(w->done);
      pthread_barrier_destroy(w->reset);

      for(t=0; t<correlation_params->threads-1; t++) {
	w = correlation_params->thread_array[t];

	if (w->thread) pthread_cancel(w->thread);

	if (w->tab_tab_produits) {
	  for (i = 0; i < correlation_params->win_nb_lig+1; i++)
	    if (w->tab_tab_produits[i]) free(w->tab_tab_produits[i]);
	  free(w->tab_tab_produits);
	}
	if (w->somme_pdt_haut) free(w->somme_pdt_haut);
	if (w->somme_pdt_bas) free(w->somme_pdt_bas);
	if (w->tab_scores) free(w->tab_scores);
	if (w->tab_scores_2) free(w->tab_scores_2);
	if (w->tab_max) free(w->tab_max);

	free(w);
      }
      free(correlation_params->thread_array);
    }
#endif

  free(correlation_params);

  return;
}

/***********************************************************************
 *
 * Fonction de copie d'une structure 'SP_CORRELATION_PARAMS'
 *
 ***********************************************************************/

SP_CORRELATION_PARAMS*
copy_correlation_params (SP_CORRELATION_PARAMS* correlation_params)
{
  SP_CORRELATION_PARAMS* new_correlation_params;

  if (correlation_params == NULL)
    return (NULL);

  new_correlation_params = create_correlation_params ();

  memcpy (new_correlation_params, correlation_params, sizeof (SP_CORRELATION_PARAMS));

  return (new_correlation_params);
}

/***********************************************************************
 *
 * Affiche des infos sur une structure 'SP_CORRELATION_PARAMS'
 *
 ***********************************************************************/

int
info_correlation_params (SP_CORRELATION_PARAMS* correlation_params)
{
  switch (correlation_params->state) {
  case SP_CORRELATION_PARAMS_NOT_INITIALIZED:
    printf ("Structure 'SP_CORRELATION_PARAMS' pas initialise\n");
    break;
  case SP_CORRELATION_PARAMS_INITIALIZED:
    printf ("Structure 'SP_CORRELATION_PARAMS' :\n");

    printf ("\tType de corrlation : ");
    switch (correlation_params->type_correlation) {
    case  SP_CORRELATION_ZNCC:
      printf ("ZNCC\n");
      break;
    case  SP_CORRELATION_ZNCC_SIGMA:
      printf ("ZNCC avec carts types\n");
      break;
    case  SP_CORRELATION_CENSUS:
      printf ("CENSUS\n");
      break;
    case  SP_CORRELATION_CENSUS_SIGMA:
      printf ("CENSUS avec carts types\n");
      break;
    default:
      return stereoError(ERR_SP_CAS_IMPOSSIBLE);
      /*     break; */
      /* Unreached */
    }

    printf ("\tType d'interpolation : ");
    switch (correlation_params->type_interpolation) {
    case  SP_INTERPOLATION_DISP_NIL:
      printf ("pas d'interpolation\n");
      break;
    case SP_INTERPOLATION_DISP_PARABOLIQUE_ONLY:
      printf ("interpolation parabolique\n");
      break;
    case SP_INTERPOLATION_DISP_BILINERAIRE_ONLY:
      printf ("interpolation bi-lineaire\n");
      break;
    case SP_INTERPOLATION_DISP_BOTH:
      printf ("interpolation bi-lineaire et parabolique\n");
      break;
    default:
      return stereoError(ERR_SP_CAS_IMPOSSIBLE);
/*       break; unreached */
    }

    printf ("\tFentre de corrlation: %d lignes par %d colonnes\n",
	    correlation_params->win_nb_lig, correlation_params->win_nb_col);
    printf ("\tEcart de disparit : de %d  %d\n",
	    correlation_params->disp_min, correlation_params->disp_max);
    printf ("\tLes seuils :\n");
    printf ("\t\tcart type min sur la fentre : %f\n",
	    correlation_params->var_th);
    printf ("\t\tScore max minimum : %f\n",
	    correlation_params->score_th);
    printf ("\t\tDiff. entre le max et ses voisins : %f\n",
	    correlation_params->diff_neighbor_max_th);
    printf ("\t\tcart max entre corrlation diverse et inverse : %d\n",
	    correlation_params->inv_th);
    break;
  default:
    return stereoError(ERR_SP_CAS_IMPOSSIBLE);
/*     break; */
    /* Unreached */
  }

  return (STEREO_OK);
}

/***********************************************************************
 *
 * Fonction d'initialisation d'une structure 'SP_CORRELATION_PARAMS'
 *
 ***********************************************************************/
int
init_correlation_params (SP_CORRELATION_PARAMS*	correlation_params,
			 SP_TYPE_CORRELATION	type_correlation,
			 SP_TYPE_INTERPOLATION_DISP	type_interpolation,
			 int			win_nb_lig,
			 int			win_nb_col,
			 int			disp_min,
			 int			disp_max,
			 int			threads,
			 double			var_th,
			 double			score_th,
			 double			diff_neighbor_max_th,
			 int			inv_th)
{
  /* Test de validit des paramtres */
  if ((type_correlation != SP_CORRELATION_CENSUS_SIGMA)
      && (type_correlation != SP_CORRELATION_CENSUS)
      && (type_correlation != SP_CORRELATION_ZNCC_SIGMA)
      && (type_correlation != SP_CORRELATION_ZNCC))
    return stereoError(ERR_SP_TYPE_CORRELATION);

  if ((type_interpolation != SP_INTERPOLATION_DISP_BILINERAIRE_ONLY)
      && (type_interpolation != SP_INTERPOLATION_DISP_PARABOLIQUE_ONLY)
      && (type_interpolation != SP_INTERPOLATION_DISP_BOTH)
      && (type_interpolation != SP_INTERPOLATION_DISP_NIL))
    return stereoError(ERR_SP_TYPE_CORRELATION);

  if ((win_nb_lig <= 0) || (win_nb_col <= 0))
    return stereoError(ERR_SP_TAILLE_FENETRE_CORRELATION);
  if (((win_nb_col % 2) == 0) || ((win_nb_lig % 2) == 0))
    return stereoError(ERR_SP_TAILLE_FENETRE_CORRELATION);

  if (disp_max <= disp_min)
    return stereoError(ERR_SP_ECART_DISPARITE);

  if ((var_th < 0.0) || (score_th < 0.0) ||
      (diff_neighbor_max_th < 0.0) || (inv_th < 0))
    return stereoError(ERR_SP_SEUIL_NEGATIF);

#if SPIX_THREADS
  if (threads < 1)
    return stereoError(ERR_SP_CAS_IMPOSSIBLE);
#else
  if (threads != 1)
    return stereoError(ERR_SP_FCT_NON_IMPLEMENTEE);
#endif

  /* On positionne les champs de 'correlation_params' passs en paramtres */
  correlation_params->type_correlation = type_correlation;
  correlation_params->type_interpolation = type_interpolation;

  correlation_params->win_nb_lig = win_nb_lig;
  correlation_params->win_nb_col = win_nb_col;
  correlation_params->disp_min = disp_min;
  correlation_params->disp_max = disp_max;
  correlation_params->threads = threads;
  correlation_params->var_th = var_th;
  correlation_params->score_th = score_th;
  correlation_params->diff_neighbor_max_th = diff_neighbor_max_th;
  correlation_params->inv_th = inv_th;

#if SPIX_THREADS
  if (threads > 1) {
    pthread_barrier_t *bs, *bd, *br;
    int i;

    bs = malloc(sizeof(*bs));
    bd = malloc(sizeof(*bd));
    br = malloc(sizeof(*br));
    if (!bs || !bd || !br)
      return stereoError(ERR_SP_MALLOC_CORRELATION_PARAMS);

    if (pthread_barrier_init(bs, NULL, threads))
      return stereoError(ERR_SP_MALLOC_CORRELATION_PARAMS);
    if (pthread_barrier_init(bd, NULL, threads))
      return stereoError(ERR_SP_MALLOC_CORRELATION_PARAMS);
    if (pthread_barrier_init(br, NULL, threads))
      return stereoError(ERR_SP_MALLOC_CORRELATION_PARAMS);

    correlation_params->thread_array = malloc((threads-1) * sizeof(void *));
    if (!correlation_params->thread_array)
      return stereoError(ERR_SP_MALLOC_CORRELATION_PARAMS);

    for(i=0; i<threads-1; i++) {
      spix_worker *w = malloc(sizeof(spix_worker));
      if (!w) return stereoError(ERR_SP_MALLOC_CORRELATION_PARAMS);

      w->id = i+1;
      w->start = bs;
      w->done = bd;
      w->reset = br;
      w->correlation_params = correlation_params;
      w->source_gauche = NULL;
      w->source_droite = NULL;
      w->images_gauches = NULL;
      w->images_droites = NULL;
      w->correlation_results = NULL;
      w->tab_tab_produits = NULL;
      w->i_min = w->i_max = 0;
      w->job = NULL;
      if (pthread_create(&w->thread, NULL, spix_worker_thread, w))
	return stereoError(ERR_SP_MALLOC_CORRELATION_PARAMS);

      correlation_params->thread_array[i] = w;
      warnx("worker thread %d ready", w->id);
    }
  } else
#endif
    correlation_params->thread_array = NULL;

  correlation_params->state = SP_CORRELATION_PARAMS_INITIALIZED;

  return (STEREO_OK);
}
