/*	$LAAS$ */

/*
 * Copyright (c) 2000 LAAS/CNRS                       --  Fri Aug  4 2006
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

/* librairies standards */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Les prototypes */
#include "libstereopixel.h"

/***********************************************************************
 *
 * Fonction de cration d'une structure 'SP_IMAGES_SIZES'
 *
 ***********************************************************************/

SP_IMAGES_SIZES*
create_images_sizes () {
  SP_IMAGES_SIZES* new_images_sizes;

  new_images_sizes = (SP_IMAGES_SIZES*) malloc (sizeof (SP_IMAGES_SIZES));

  if (new_images_sizes == NULL) {
    stereoError (ERR_SP_MALLOC_IMAGES_SIZES);
    return (NULL);
  }

  new_images_sizes->state = SP_IMAGES_SIZES_NOT_INITIALIZED;

  return (new_images_sizes);
}

/***********************************************************************
 *
 * Fonction de destruction d'une structure 'SP_IMAGES_SIZES'
 *
 ***********************************************************************/

void
destroy_images_sizes (SP_IMAGES_SIZES* images_sizes)
{
  if (images_sizes == NULL)
    return;

  sc_clean_images_sizes(images_sizes);
  free (images_sizes);

  return;
}

void
sc_clean_images_sizes(SP_IMAGES_SIZES *images_sizes)
{
   if (images_sizes != NULL)
      images_sizes->state = SP_IMAGES_SIZES_NOT_INITIALIZED;
}

/***********************************************************************
 *
 * Fonction de copie d'une structure 'SP_IMAGES_SIZES'
 *
 ***********************************************************************/

SP_IMAGES_SIZES*
copy_images_sizes (SP_IMAGES_SIZES* images_sizes)
{
  SP_IMAGES_SIZES* new_images_sizes;

  if (images_sizes == NULL)
    return (NULL);

  new_images_sizes = create_images_sizes ();

  memcpy (new_images_sizes, images_sizes, sizeof (SP_IMAGES_SIZES));

  return (new_images_sizes);
}

/***********************************************************************
 *
 * Affiche des infos sur une structure 'SP_IMAGES_SIZES'
 *
 ***********************************************************************/

int
info_images_sizes (SP_IMAGES_SIZES* images_sizes)
{
  switch (images_sizes->state) {
  case SP_IMAGES_SIZES_NOT_INITIALIZED:
    printf ("Structure 'SP_IMAGES_SIZES' pas initialise\n");
    break;
  case SP_IMAGES_SIZES_INITIALIZED:
    printf ("Structure 'SP_IMAGES_SIZES' :\n");
    printf ("\tTaille originale %d lignes par %d colonnes\n",
	    images_sizes->nb_lig_original, images_sizes->nb_col_original);
    printf ("\tTaux de rduction appliqu pralablement  tout traitement : %d\n",
	    images_sizes->previousReduction);
    printf ("\t\tsuivant les lignes : %d\n", images_sizes->taux_reduction_i);
    printf ("\tTaux de rduction appliqu pour traitement :\n");
    printf ("\t\tsuivant les lignes : %d\n", images_sizes->taux_reduction_i);
    printf ("\t\tsuivant les colonnes : %d\n", images_sizes->taux_reduction_j);
    printf ("\t\t -> taille rduite %d lignes par %d colonnes\n",
	    images_sizes->nb_lig_reduit, images_sizes->nb_col_reduit);
    printf ("\tSous image extraite : %.2f < lignes < %.2f, %.2f < colonnes < %.2f\n",
	    images_sizes->debut_ligne, images_sizes->fin_ligne,
	    images_sizes->debut_colonne, images_sizes->fin_colonne);
    printf ("\tCoordonnes dans l'image originale :\n");
    printf ("\t\tde %d  %d lignes, de %d  %d colonnes\n",
	    images_sizes->i_min_orig, images_sizes->i_max_orig,
	    images_sizes->j_min_orig, images_sizes->j_max_orig);
    printf ("\tCoordonnes dans l'image rduite :\n");
    printf ("\t\tde %d  %d lignes, de %d  %d colonnes\n",
	    images_sizes->i_min, images_sizes->i_max,
	    images_sizes->j_min, images_sizes->j_max);
    printf ("\tTaille finale %d lignes par %d colonnes\n",
	    images_sizes->nb_lig_final, images_sizes->nb_col_final);
    break;
  default:
    return stereoError(ERR_SP_CAS_IMPOSSIBLE);
/*     break; */
    /* Unreached */
  }

  return (STEREO_OK);
}

/***********************************************************************
 *
 * Fonction d'initialisation d'une structure 'SP_IMAGES_SIZES'
 *
 ***********************************************************************/
int
init_images_sizes (SP_IMAGES_SIZES*	images_sizes,
		   int			nblig,
		   int			nbcol,
		   int                  previousReduction,
		   int			taux_reduction_i,
		   int			taux_reduction_j,
		   double		debut_ligne,
		   double		fin_ligne,
		   double		debut_colonne,
		   double		fin_colonne)
{
  int	i_min_crop, i_max_crop;
  int	j_min_crop, j_max_crop;

  /* Test de validit des paramtres */
  if ((nblig <= 0) || (nbcol <= 0)) {
    return stereoError(ERR_SP_IMAGES_SIZES_NEGATIVES); 
  }

/*     return error(ERR_SP_IMAGES_SIZES_NEGATIVES); */

  if ((taux_reduction_i <= 0) || (taux_reduction_j <= 0) || (previousReduction <= 0))
      return stereoError(ERR_SP_TAUX_REDUCTION_NEGATIF);

  if (taux_reduction_i != taux_reduction_j)
    return stereoError(ERR_SP_FCT_NON_IMPLEMENTEE);

  if ((debut_ligne < 0.0) || (debut_ligne >= 1.0)
      || (fin_ligne > 1.0) || (fin_ligne <= 0.0)
      || (debut_colonne < 0.0) || (debut_colonne >= 1.0)
      || (fin_colonne > 1.0) || (fin_colonne <= 0.0)
      || (fin_ligne <= debut_ligne)
      || (fin_colonne <= debut_colonne))
    return stereoError(ERR_SP_IMAGES_SIZES_SOUS_IMAGE);

  /* On positionne les champs de 'images_sizes' passs en paramtres */
  images_sizes->nb_lig_original = nblig;
  images_sizes->nb_col_original = nbcol;
  images_sizes->taille_originale = nblig * nbcol;
  images_sizes->previousReduction = previousReduction;
  images_sizes->taux_reduction_i = taux_reduction_i;
  images_sizes->taux_reduction_j = taux_reduction_j;
  images_sizes->debut_ligne = debut_ligne;
  images_sizes->fin_ligne = fin_ligne;
  images_sizes->debut_colonne = debut_colonne;
  images_sizes->fin_colonne = fin_colonne;

  /* On dtermine tous les champs de 'images_sizes' : les tailles de l'image
     rduite (qui doivent tre paires) ... */
  images_sizes->nb_lig_reduit
    = (nblig / taux_reduction_i) ; //- ((nblig / taux_reduction_i) % 2);
  images_sizes->nb_col_reduit
    = (nbcol / taux_reduction_j) ; //- ((nbcol / taux_reduction_j) % 2);
  images_sizes->taille_reduite
    = (images_sizes->nb_lig_reduit) * (images_sizes->nb_col_reduit);

  /* ... les coordonnes de la sous-image dans l'image rduite... */
  images_sizes->i_min = (int) (debut_ligne * images_sizes->nb_lig_reduit);
  images_sizes->i_max = (int) (fin_ligne * images_sizes->nb_lig_reduit);
  images_sizes->j_min = (int) (debut_colonne * images_sizes->nb_col_reduit);
  images_sizes->j_max = (int) (fin_colonne * images_sizes->nb_col_reduit);


  /* ... les coordonnes de la sous-image dans l'image originale... */
  images_sizes->i_min_orig = taux_reduction_i * images_sizes->i_min;
  images_sizes->i_max_orig = taux_reduction_i * images_sizes->i_max;
  images_sizes->j_min_orig = taux_reduction_j * images_sizes->j_min;
  images_sizes->j_max_orig = taux_reduction_j * images_sizes->j_max;

  /* ... les tailles finales (qui doivent tre paires)... */
/*   if (((images_sizes->i_max - images_sizes->i_min) % 2) != 0) */
/*     images_sizes->i_max -= 1; */
  images_sizes->nb_lig_final = images_sizes->i_max - images_sizes->i_min;
/*   if (((images_sizes->j_max - images_sizes->j_min) % 2) != 0) */
/*     images_sizes->j_max -= 1; */
  images_sizes->nb_col_final = images_sizes->j_max - images_sizes->j_min;

  images_sizes->taille_finale
    = (images_sizes->nb_lig_final) * (images_sizes->nb_col_final);

  /* ... et les coordonnes de l'image croppe dans l'image originale. */
  i_min_crop = (int) (debut_ligne * images_sizes->nb_lig_original);
  i_max_crop = (int) (fin_ligne * images_sizes->nb_lig_original);
  j_min_crop = (int) (debut_colonne * images_sizes->nb_col_original);
  j_max_crop = (int) (fin_colonne * images_sizes->nb_col_original);

  images_sizes->nb_lig_crop = i_max_crop - i_min_crop;
  images_sizes->nb_col_crop = j_max_crop - j_min_crop;
  images_sizes->taille_crop = images_sizes->nb_lig_crop * images_sizes->nb_col_crop;


  /* On vrifie qu'on ne tombe pas sur une image de taille nulle */
  if ((images_sizes->nb_lig_final == 0) || (images_sizes->nb_col_final == 0))
    return stereoError(ERR_SP_IMAGES_SIZES_SOUS_IMAGE);

  images_sizes->state = SP_IMAGES_SIZES_INITIALIZED;

  return (STEREO_OK);
}
