/*	$LAAS$ */

/*
 * Copyright (c) 2000 LAAS/CNRS                       --  Fri Aug  4 2006
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

/* librairies standards */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Les erreurs, structures et numrations locales */
#include "libstereopixel.h"

#define FACTEUR_DISTORTION 1e6


/***********************************************************************
 *
 * Fonction de cration d'une structure 'SP_RECTIF_PARAMS'
 *
 ***********************************************************************/

SP_RECTIF_PARAMS*
create_rectif_params () {
  SP_RECTIF_PARAMS* new_rectif_params;

  new_rectif_params
    = (SP_RECTIF_PARAMS*) malloc (sizeof (SP_RECTIF_PARAMS));

  if (new_rectif_params == NULL) {
    stereoError (ERR_SP_MALLOC_RECTIF_PARAMS);
    return (NULL);
  }

  new_rectif_params->table_i = NULL;
  new_rectif_params->table_j = NULL;
  new_rectif_params->inverse_table_i = NULL;
  new_rectif_params->inverse_table_j = NULL;

  new_rectif_params->state = SP_RECTIF_PARAMS_NOT_INITIALIZED;

  return (new_rectif_params);
}

/***********************************************************************
 *
 * Fonction de destruction d'une structure 'SP_RECTIF_PARAMS'
 *
 ***********************************************************************/

void
destroy_rectif_params (SP_RECTIF_PARAMS* rectif_params)
{
  if (rectif_params == NULL)
    return;

  if (rectif_params->inverse_table_i != NULL)
    free (rectif_params->inverse_table_i);
  if (rectif_params->inverse_table_j != NULL)
    free (rectif_params->inverse_table_j);

  if (rectif_params->table_i != NULL)
    free (rectif_params->table_i);
  if (rectif_params->table_j != NULL)
    free (rectif_params->table_j);

  free (rectif_params);

  return;
}

/***********************************************************************
 *
 * Fonction de copie d'une structure 'SP_RECTIF_PARAMS'
 *
 ***********************************************************************/

SP_RECTIF_PARAMS*
copy_rectif_params (SP_RECTIF_PARAMS* rectif_params)
{
  SP_RECTIF_PARAMS* new_rectif_params;

  if (rectif_params == NULL)
    return (NULL);

  new_rectif_params = create_rectif_params ();

  memcpy (new_rectif_params, rectif_params, sizeof (SP_RECTIF_PARAMS));

  if (rectif_params->table_i != NULL) {
    printf ("ERR_SP_FCT_NON_IMPLEMENTEE");
    return (NULL);
  }

  if (rectif_params->table_j != NULL) {
    printf ("ERR_SP_FCT_NON_IMPLEMENTEE");
    return (NULL);
  }

  if (rectif_params->inverse_table_i != NULL) {
    printf ("ERR_SP_FCT_NON_IMPLEMENTEE");
    return (NULL);
  }

  if (rectif_params->inverse_table_j != NULL) {
    printf ("ERR_SP_FCT_NON_IMPLEMENTEE");
    return (NULL);
  }

  return (new_rectif_params);
}

/***********************************************************************
 *
 * Affiche des infos sur une structure 'SP_RECTIF_PARAMS'
 *
 ***********************************************************************/

int
info_rectif_params (SP_RECTIF_PARAMS* rectif_params)
{
  switch (rectif_params->state) {
  case SP_RECTIF_PARAMS_NOT_INITIALIZED:
    printf ("Structure 'SP_RECTIF_PARAMS' pas initialise\n");
    break;
  case SP_RECTIF_PARAMS_MATRICES_INITIALIZED:
    printf ("Structure 'SP_RECTIF_PARAMS' (matrices seules):\n");

    printf ("\tMatrice intrinsque:\n");
    printf ("\t%f \t%f \t%f \t%f\n",
	    rectif_params->intrinseque[0], rectif_params->intrinseque[1],
	    rectif_params->intrinseque[2], rectif_params->intrinseque[3]);
    printf ("\t%f \t%f \t%f \t%f\n",
	    rectif_params->intrinseque[4], rectif_params->intrinseque[5],
	    rectif_params->intrinseque[6], rectif_params->intrinseque[7]);
    printf ("\t%f \t%f \t%f \t%f\n",
	    rectif_params->intrinseque[8], rectif_params->intrinseque[9],
	    rectif_params->intrinseque[10], rectif_params->intrinseque[11]);
    printf ("\t%f \t%f \t%f \t%f\n",
	    rectif_params->intrinseque[12], rectif_params->intrinseque[13],
	    rectif_params->intrinseque[14], rectif_params->intrinseque[15]);

    printf ("\tMatrice de rectification:\n");
    printf ("\t%f \t%f \t%f\n",
	    rectif_params->rectification[0], rectif_params->rectification[1],
	    rectif_params->rectification[2]);
    printf ("\t%f \t%f \t%f\n",
	    rectif_params->rectification[3], rectif_params->rectification[4],
	    rectif_params->rectification[5]);
    printf ("\t%f \t%f \t%f\n",
	    rectif_params->rectification[6], rectif_params->rectification[7],
	    rectif_params->rectification[8]);

    printf ("\tMatrice intrinsque rectifie:\n");
    printf ("\t%f \t%f \t%f \t%f\n",
	    rectif_params->intrinseque_rectifiee[0], rectif_params->intrinseque_rectifiee[1],
	    rectif_params->intrinseque_rectifiee[2], rectif_params->intrinseque_rectifiee[3]);
    printf ("\t%f \t%f \t%f \t%f\n",
	    rectif_params->intrinseque_rectifiee[4], rectif_params->intrinseque_rectifiee[5],
	    rectif_params->intrinseque_rectifiee[6], rectif_params->intrinseque_rectifiee[7]);
    printf ("\t%f \t%f \t%f \t%f\n",
	    rectif_params->intrinseque_rectifiee[8], rectif_params->intrinseque_rectifiee[9],
	    rectif_params->intrinseque_rectifiee[10], rectif_params->intrinseque_rectifiee[11]);
    printf ("\t%f \t%f \t%f \t%f\n",
	    rectif_params->intrinseque_rectifiee[12], rectif_params->intrinseque_rectifiee[13],
	    rectif_params->intrinseque_rectifiee[14], rectif_params->intrinseque_rectifiee[15]);

    printf ("\tParamtres de correction de la distortion:\n");
    printf ("\t%f\n", rectif_params->distortion[0]);

    if ((rectif_params->table_i != NULL) & (rectif_params->table_j != NULL))
      printf ("Tables de rectification directes dtermines\n");

    if ((rectif_params->inverse_table_i != NULL)
	& (rectif_params->inverse_table_j != NULL))
      printf ("Tables de rectification inverses dtermines\n");
    break;


  default:
    return stereoError(ERR_SP_CAS_IMPOSSIBLE);
/*     break; */
    /* Unreached */
  }

  return (STEREO_OK);
}

/***********************************************************************
 *
 * Fonction d'initialisation d'une structure 'SP_RECTIF_PARAMS'
 *
 ***********************************************************************/
int
init_rectif_params (SP_RECTIF_PARAMS*	rectif_params,
		    SP_IMAGES_SIZES*	images_sizes,
		    double*		mat_intrinseque,
		    double*		mat_rectification,
		    double*		mat_distortion,
		    double*		mat_intrinseque_rectifiee)

  {
  /* Test de validit des paramtres */
  if ((mat_intrinseque == NULL)
      || (mat_rectification == NULL)
      || (mat_distortion == NULL)
      || (mat_intrinseque_rectifiee == NULL))
    return stereoError(ERR_SP_MATRICE_NULLE);

  if (images_sizes->state != SP_IMAGES_SIZES_INITIALIZED)
    return stereoError(ERR_SP_IMAGES_SIZES_NOT_INITIALIZED);

  /* On copie toutes les matrices. !!! XXX On suppose qu'elles ont bien les
     tailles spcifies dans la structure */
  memcpy (rectif_params->intrinseque, mat_intrinseque, 16 * sizeof (double));
  memcpy (rectif_params->rectification, mat_rectification, 9 * sizeof (double));
  memcpy (rectif_params->distortion, mat_distortion, 3 * sizeof (double));
  memcpy (rectif_params->intrinseque_rectifiee, mat_intrinseque_rectifiee,
	  16 * sizeof (double));

  /* Copie de la structure 'images_sizes' dans le champ adquat de la
     structure rectif_params */
  memcpy (&(rectif_params->images_sizes), images_sizes, sizeof (SP_IMAGES_SIZES));

  rectif_params->state = SP_RECTIF_PARAMS_MATRICES_INITIALIZED;

  return (STEREO_OK);
}

/***********************************************************************
 *
 * Fonction de calcul des tables de rectification d'une structure
 * 'SP_RECTIF_PARAMS' (l'origine du code est dans la fct 'rectifie_sous_ima1',
 * dans le fichier '../PreProcessing/rectification.c').
 *
 ***********************************************************************/
int
init_rectif_params_tables (SP_CORRECTION_DISTORTION	type_distortion,
			   SP_RECTIF_PARAMS*		rectif_params)

{
  int		nblig;
  int		nbcol;

  int		i,j;		/* Indices d'un pixel de l'image rectifiee
				   (donc des tables de rectification) */

  double	u,v, r;		/* Indices du pixel correspondant dans
				 * l'image d'origine */

  short*	table_i_ptr;
  short*	table_j_ptr;	/* Pointeurs dans les tables de rectification */

  double	rect_copy[9];	/* Copie de travail de 'rectification'
				   (dans 'rectif_params') */

  double	r0, r1, r2;	/* Pour dterminer la projection */

  double	i_centre, j_centre;
				/* Centre de l'image source (lue dans
				   'rectification') */

  int		nb_out = 0;	/* Juste pour voir */

  double	k = 0.0;
  double	lambda = 0.0;	/* Init. pr viter un warning */
  double	coefficient;	/* Ces 3 variables sont l pour
				   prendre en compte la distorsion */

  double	alpha_u;
  double	alpha_v;	/* Facteurs d'echelle de la matrice
				   intrinsque  */

  /*******************************************************************/

  /* Test de validit des paramtres */
  if ((rectif_params == NULL)
      || (rectif_params->state != SP_RECTIF_PARAMS_MATRICES_INITIALIZED))
    return stereoError(ERR_SP_RECTIF_PARAMS_STATE);

  /* Allocation des tables de rectification directe */
  nblig = rectif_params->images_sizes.nb_lig_original;
  nbcol = rectif_params->images_sizes.nb_col_original;

  alpha_u = rectif_params->intrinseque[0];
  alpha_v = rectif_params->intrinseque[5];

  rectif_params->table_i = (short*) malloc (nblig * nbcol * sizeof (short));
  if (rectif_params->table_i == NULL)
    return stereoError (ERR_SP_MALLOC_RECTIF_TABLE);
  rectif_params->table_j = (short*) malloc (nblig * nbcol * sizeof (short));
  if (rectif_params->table_i == NULL)
    return stereoError (ERR_SP_MALLOC_RECTIF_TABLE);

  /* Mise  '-1' des lments des tables. TODO : Tester le memset */
/*   memset ( (short*) rectif_params->table_i, -1, nblig * nbcol * sizeof (short)); */
/*   memset ( (short*) rectif_params->table_j, -1, nblig * nbcol * sizeof (short));/ */

  table_i_ptr = rectif_params->table_i;
  table_j_ptr = rectif_params->table_j;
  for (i = 0; i < nblig; i++) {
    for (j = 0; j < nbcol; j++) {
      *(table_i_ptr) = -1;
      *(table_j_ptr) = -1;
      table_i_ptr++;
      table_j_ptr++;
    }
  }


  /* Calcul des tables de rectification directe */

  /* On rcupre la position du centre de l'image dans la matrice de
     calibration intrinsque */
  i_centre =  rectif_params->intrinseque[2];
  j_centre =  rectif_params->intrinseque[6];

  /* Prise en compte de la distortion. Voir la thse de Jean-Jos
     ORTEU page 89 (!) */
  switch (type_distortion) {
  case SP_DISTORTION_NIL:
    ;
    break;

  case SP_DISTORTION_R_OLD:
    k = (rectif_params->distortion[0]) / FACTEUR_DISTORTION;
    lambda = (alpha_u * alpha_u) / (alpha_v * alpha_v);
    break;

  case SP_DISTORTION_R:
    k = rectif_params->distortion[0]; /* Pas de facteur le mutiplication
					 FACTEUR_DISTORTION ??? */
    break;

  case SP_DISTORTION_R3_T2:
    ;
    break;

  default:
    return stereoError(ERR_SP_CAS_IMPOSSIBLE);
    /* Unreached */
    /*     break; */
  }

  /* Changement d'echelle de la matrice de rectification. */
  memcpy(rect_copy, rectif_params->rectification, sizeof(rect_copy));
#if 0
  /* XXX what's that??? */
  rect_copy [2] = rect_copy [2];
  rect_copy [5] = rect_copy [5];
  rect_copy [6] = rect_copy [6];
  rect_copy [7] = rect_copy [7];
#endif

  /* On se positionne  l'origine dans les tables d'indices. ATTENTION !!! :
     on les range comme des ima1, bien qu'il s'agisse d'ima4. Donc, gaffe. */
  table_i_ptr = rectif_params->table_i;
  table_j_ptr = rectif_params->table_j;

  /* On boucle sur les pixels des tables (donc de l'image rectifie) */
  for (i = 0; i < nblig; i++) {
    for (j = 0; j < nbcol; j++) {

      /* On dtermine la position du pixel correspondant dans l'image
	 originale (projection linaire) */
      r0 = (rect_copy[0] * i) + (rect_copy[1] * j) + rect_copy[2];
      r1 = (rect_copy[3] * i) + (rect_copy[4] * j) + rect_copy[5];
      r2 = (rect_copy[6] * i) + (rect_copy[7] * j) + rect_copy[8];

      u = r0 / r2;
      v = r1 / r2;

      /* Correction de la distortion */
      switch (type_distortion) {
      case SP_DISTORTION_NIL:
	;
	break;
      case SP_DISTORTION_R_OLD:
	/* on corrige la distortion */
        u -= i_centre;
	v -= j_centre;
	coefficient = 1.0 + k * ( (u * u) + lambda * (v * v));
	u = i_centre + (u * coefficient);
	v = j_centre + (v * coefficient);
	break;
      case SP_DISTORTION_R:
	/* on corrige la distortion (nouveau modle en R) */
        u = (u - i_centre) / alpha_u;
	v = (v - j_centre) / alpha_v;
	coefficient = 1.0 + k * ( (u * u) + ( v * v ));
	u = i_centre + (u * coefficient * alpha_u);
	v = j_centre + (v * coefficient * alpha_v);
	break;
      case SP_DISTORTION_R3_T2:
        u = (u - i_centre) / alpha_u;
	v = (v - j_centre) / alpha_v;
	r = (u * u) + (v * v);
	coefficient = 1.0 + (r * rectif_params->distortion[0])
		      + (r * r * rectif_params->distortion[1])
		      + (r * r * r * rectif_params->distortion[2]);
	u = i_centre + (u * coefficient * alpha_u);
	v = j_centre + (v * coefficient * alpha_v);

	break;

      default:
	return stereoError (ERR_SP_CAS_IMPOSSIBLE);
	/* 	break; */
      }

      /* Test si le point trouv est hors du cadre. TODO : Attention : il y a
	 ici une diffrence avec la fct de rectification. Pourquoi ?  cause
	 du '+ 0.5' ? */
      if ((u < 0) || (v < 0) || (u > (nblig - 1)) || (v > (nbcol - 1))) {
 	nb_out++;
	*table_i_ptr = - 1;
	*table_j_ptr = - 1;
      }
      else { /* On rempli les tables. Bien entendu, pas d'interpolation
		possible */
	*table_i_ptr = (short) (u + 0.5);
	*table_j_ptr = (short) (v + 0.5);
      }
      table_i_ptr++;
      table_j_ptr++;
    }
  }

  /* C'est fini. */
  printf ("\t%d pixels hors du cadre\n", nb_out);

  return (STEREO_OK);
}

/***********************************************************************
 *
 * Fonction de calcul des tables de rectification inaverses d'une structure
 * 'SP_RECTIF_PARAMS',  partir des tables de rectification directes dj
 * dtermines.
 *
 ***********************************************************************/
int
init_rectif_params_inverse_tables (SP_RECTIF_PARAMS*	rectif_params)

{
  int		i, j;
  int		u, v;

  int		nblig;
  int		nbcol;

  short*	table_i_ptr;
  short*	table_j_ptr;

  short*	inverse_table_i_ptr;
  short*	inverse_table_j_ptr;

  int		nb_out = 0;

  /* Test de validit des paramtres */
  if ((rectif_params->state != SP_RECTIF_PARAMS_MATRICES_INITIALIZED)
      || (rectif_params->table_i == NULL)
      || (rectif_params->table_j == NULL))
    return stereoError(ERR_SP_RECTIF_PARAMS_STATE);

  /* Allocation des tables de rectification inverse */
  nblig = rectif_params->images_sizes.nb_lig_original;
  nbcol = rectif_params->images_sizes.nb_col_original;

  rectif_params->inverse_table_i = (short*) malloc (nblig * nbcol * sizeof (short));
  if (rectif_params->inverse_table_i == NULL)
    return stereoError (ERR_SP_MALLOC_RECTIF_TABLE);
  rectif_params->inverse_table_j = (short*) malloc (nblig * nbcol * sizeof (short));
  if (rectif_params->inverse_table_i == NULL)
    return stereoError (ERR_SP_MALLOC_RECTIF_TABLE);

  /* Mise  '-1' des lments des tables. TODO : Tester le memset */
/*   memset ( (short*) rectif_params->table_i, -1, nblig * nbcol * sizeof (short)); */
/*   memset ( (short*) rectif_params->table_j, -1, nblig * nbcol * sizeof (short));/ */
  inverse_table_i_ptr = rectif_params->inverse_table_i;
  inverse_table_j_ptr = rectif_params->inverse_table_j;
  for (i = 0; i < nblig; i++) {
    for (j = 0; j < nbcol; j++) {
      *(inverse_table_i_ptr) = -1;
      *(inverse_table_j_ptr) = -1;
      inverse_table_i_ptr++;
      inverse_table_j_ptr++;
    }
  }

  /* Calcul des tables de rectification inverse */
  /* On se positionne  l'origine dans les images d'indices. ATTENTION !!! :
     on  le range comme des ima1, bien qu'il s'agisse d'ima4. De toutes
     faons, on s'en fout, non ? */
  table_i_ptr = rectif_params->table_i;
  table_j_ptr = rectif_params->table_j;

  for (i = 0; i < nblig; i++) {
    for (j = 0; j < nbcol; j++) {

      u = (int) *table_i_ptr;
      if (u < 0)
	nb_out++;
      else {
	v = (int) *table_j_ptr;
	*(rectif_params->inverse_table_i + (u * nbcol) + v) = i;
	*(rectif_params->inverse_table_j + (u * nbcol) + v) = j;
      }

      table_i_ptr++;
      table_j_ptr++;
    }
  }

  /* On bouche les trous */
  fill_inverse_tables (rectif_params);

/*   printf ("\t%d pixels hors du cadre\n", nb_out); */

  return (STEREO_OK);
}

/********************************************************************
 *
 * Voici une fonction compltement ad-hoc qui bouche les trous des images
 * d'incides de rectification inverse (trous dus aux erreurs d'arrondi ?)
 *
 ********************************************************************/
int
fill_inverse_tables (SP_RECTIF_PARAMS*	rectif_params)


{
  int	nblig;
  int	nbcol;
  int	i, j;

  int	pix_haut;
  int	pix;
  int	pix_bas;

  short*	inverse_table_i_ptr;
  short*	inverse_table_j_ptr; /* Pointeurs sur les images d'indices */


  /* On se positionne  l'origine dans les images d'indices. ATTENTION !!! :
     cette fois on se promne dedans en les considrant bien comme des ima4 :
     on traite les images colonnes par colonnes (mais pourquoi cela au fait ?)
  */
  nblig = rectif_params->images_sizes.nb_lig_original;
  nbcol = rectif_params->images_sizes.nb_col_original;

  for (j = 0; j < nblig; j++) {
    /* On se positionne  un pixel du haut de la colonne */
    inverse_table_i_ptr = rectif_params->inverse_table_i + (j * nblig) + 1;
    inverse_table_j_ptr = rectif_params->inverse_table_j + (j * nblig) + 1;

    for (i = 1; i < nbcol - 1; i++) {

      pix_haut = *(inverse_table_i_ptr - 1);
      pix = *(inverse_table_i_ptr);
      pix_bas = *(inverse_table_i_ptr + 1);

      if ((pix < 0)
	  && (pix_bas > 0)
	  && (pix_haut > 0))
	*inverse_table_i_ptr = pix_bas;

      pix_haut = *(inverse_table_j_ptr - 1);
      pix = *(inverse_table_j_ptr);
      pix_bas = *(inverse_table_j_ptr + 1);

      if ((pix < 0)
	  && (pix_bas > 0)
	  && (pix_haut > 0))
	*inverse_table_j_ptr = pix_bas;

      inverse_table_i_ptr++;
      inverse_table_j_ptr++;
    }
  }

  /* Tout s'est bien pass */
  return (STEREO_OK);
}
